# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from collections import OrderedDict
from pathlib import Path
from typing import List

import torch
from torch import Tensor

from nncf.common.logging import nncf_logger
from nncf.common.utils.decorators import skip_if_dependency_unavailable
from nncf.common.utils.dot_file_rw import write_dot_graph
from nncf.torch.graph.transformations.commands import ExtraCompressionModuleType
from nncf.torch.nncf_network import NNCFNetwork
from nncf.torch.quantization.adjust_padding import add_adjust_padding_nodes
from nncf.torch.quantization.layers import QUANTIZATION_MODULES
from nncf.torch.quantization.precision_init.adjacent_quantizers import GroupsOfAdjacentQuantizers
from nncf.torch.quantization.precision_init.definitions import QConfigSequenceForHAWQToEvaluate
from nncf.torch.quantization.precision_init.perturbations import PerturbationObserver
from nncf.torch.quantization.precision_init.perturbations import Perturbations
from nncf.torch.quantization.precision_init.traces_order import TracesPerLayer
from nncf.torch.utils import get_all_modules_by_type


class HAWQDebugger:
    def __init__(
        self,
        weight_qconfig_sequences_in_trace_order: List["QConfigSequenceForHAWQToEvaluate"],
        perturbations: Perturbations,
        weight_observers_for_each_covering_configuration: List[List[PerturbationObserver]],
        traces_per_layer: TracesPerLayer,
        bitwidths: List[int],
    ):
        self._weight_qconfig_sequences_in_trace_order = weight_qconfig_sequences_in_trace_order
        self._num_weights = len(traces_per_layer.traces_order)
        self._perturbations = perturbations

        from nncf.common.utils.debug import DEBUG_LOG_DIR

        self._dump_dir = Path(DEBUG_LOG_DIR) / Path("hawq_dumps")
        self._dump_dir.mkdir(parents=True, exist_ok=True)

        self._traces_order = traces_per_layer.traces_order
        self._traces_per_layer = traces_per_layer.get_all()

        num_of_weights = []
        norm_of_weights = []
        for i in range(self._num_weights):
            trace_index = self._traces_order.get_execution_index_by_traces_index(i)
            num_of_weights.append(weight_observers_for_each_covering_configuration[0][trace_index].get_numels())
            norm_of_weights.append(weight_observers_for_each_covering_configuration[0][trace_index].get_input_norm())
        self._num_weights_per_layer = torch.Tensor(num_of_weights)
        self._norm_weights_per_layer = torch.Tensor(norm_of_weights)

        bits_in_megabyte = 2**23
        self._model_sizes = []
        for qconfig_sequence in self._weight_qconfig_sequences_in_trace_order:
            size = (
                torch.sum(
                    torch.Tensor([qconfig.num_bits for qconfig in qconfig_sequence]) * self._num_weights_per_layer
                ).item()
                / bits_in_megabyte
            )
            self._model_sizes.append(size)
        self._bitwidths = bitwidths

    @staticmethod
    def get_all_quantizers_per_full_scope(model):
        all_quantizations = OrderedDict()
        for class_type in QUANTIZATION_MODULES.registry_dict.values():
            quantization_type = class_type.__name__
            all_quantizations.update(
                get_all_modules_by_type(
                    model.nncf.get_compression_modules_by_type(ExtraCompressionModuleType.EXTERNAL_QUANTIZER),
                    quantization_type,
                )
            )
            all_quantizations.update(get_all_modules_by_type(model, quantization_type))
        all_quantizations = OrderedDict(sorted(all_quantizations.items(), key=lambda x: str(x[0])))
        return all_quantizations

    @skip_if_dependency_unavailable(dependencies=["matplotlib.pyplot"])
    def dump_avg_traces(self):
        import matplotlib.pyplot as plt

        dump_file = os.path.join(self._dump_dir, "avg_traces_per_layer")
        torch.save(self._traces_per_layer, dump_file)
        fig = plt.figure()
        fig.suptitle("Average Hessian Trace")
        ax = fig.add_subplot(2, 1, 1)
        ax.set_yscale("log")
        ax.set_xlabel("weight quantizers")
        ax.set_ylabel("average hessian trace")
        ax.plot(self._traces_per_layer.cpu().numpy())
        plt.savefig(dump_file)

    @skip_if_dependency_unavailable(dependencies=["matplotlib.pyplot"])
    def dump_metric_MB(self, metric_per_qconfig_sequence: List[Tensor]):
        import matplotlib.pyplot as plt

        list_to_plot = [cm.item() for cm in metric_per_qconfig_sequence]
        fig = plt.figure()
        fig.suptitle("Pareto Frontier")
        ax = fig.add_subplot(2, 1, 1)
        ax.set_yscale("log")
        ax.set_xlabel("Model Size (MB)")
        ax.set_ylabel("Metric value (total perturbation)")
        ax.scatter(self._model_sizes, list_to_plot, s=20, facecolors="none", edgecolors="r")
        cm = torch.Tensor(metric_per_qconfig_sequence)
        cm_m = cm.median().item()
        qconfig_index = metric_per_qconfig_sequence.index(cm_m)
        ms_m = self._model_sizes[qconfig_index]
        ax.scatter(ms_m, cm_m, s=30, facecolors="none", edgecolors="b", label="median from all metrics")
        ax.legend()
        plt.savefig(os.path.join(self._dump_dir, "Pareto_Frontier"))
        nncf_logger.debug(
            f"Distribution of HAWQ metrics: "
            f"min_value={cm.min().item():.3f}, "
            f"max_value={cm.max().item():.3f}, "
            f"median_value={cm_m:.3f}, "
            f"median_index={qconfig_index}, "
            f"total_number={len(metric_per_qconfig_sequence)}"
        )

    @skip_if_dependency_unavailable(dependencies=["matplotlib.pyplot"])
    def dump_metric_flops(
        self, metric_per_qconfig_sequence: List[Tensor], flops_per_config: List[float], choosen_qconfig_index: int
    ):
        import matplotlib.pyplot as plt

        list_to_plot = [cm.item() for cm in metric_per_qconfig_sequence]
        fig = plt.figure()
        fig.suptitle("Pareto Frontier")
        ax = fig.add_subplot(1, 1, 1)
        ax.set_xlabel("Compression ratio: total INT8 Bits Complexity / total MIXED INT Bits Complexity")
        ax.set_ylabel("Metric value (total perturbation)")
        ax.scatter(flops_per_config, list_to_plot, s=10, alpha=0.3)  # s=20, facecolors='none', edgecolors='r')
        flops_per_config = [torch.Tensor([v]) for v in flops_per_config]
        cm = torch.Tensor(flops_per_config)
        cm_m = cm.median().item()
        configuration_index = flops_per_config.index(cm_m)
        ms_m = metric_per_qconfig_sequence[configuration_index].item()
        ax.scatter(cm_m, ms_m, s=30, facecolors="none", edgecolors="b", label="median from all metrics")
        cm_c = metric_per_qconfig_sequence[choosen_qconfig_index].item()
        fpc_c = flops_per_config[choosen_qconfig_index].item()
        ax.scatter(fpc_c, cm_c, s=30, facecolors="none", edgecolors="r", label="chosen config")

        ax.legend()
        plt.savefig(os.path.join(self._dump_dir, "Pareto_Frontier_compress_ratio"))

    @skip_if_dependency_unavailable(dependencies=["matplotlib.pyplot"])
    def dump_density_of_quantization_noise(self):
        noise_per_config: List[Tensor] = []
        for qconfig_sequence in self._weight_qconfig_sequences_in_trace_order:
            qnoise = 0
            for i in range(self._num_weights):
                execution_index = self._traces_order.get_execution_index_by_traces_index(i)
                qnoise += self._perturbations.get(layer_id=execution_index, qconfig=qconfig_sequence[i])
            noise_per_config.append(qnoise)

        list_to_plot = [cm.item() for cm in noise_per_config]
        import matplotlib.pyplot as plt

        fig = plt.figure()
        fig.suptitle("Density of quantization noise")
        ax = fig.add_subplot(2, 1, 1)
        ax.set_yscale("log")
        ax.set_xlabel("Blocks")
        ax.set_ylabel("Noise value")
        ax.scatter(self._model_sizes, list_to_plot, s=20, alpha=0.3)
        ax.legend()
        plt.savefig(os.path.join(self._dump_dir, "Density_of_quantization_noise"))

    @skip_if_dependency_unavailable(dependencies=["matplotlib.pyplot"])
    def dump_perturbations_ratio(self):
        import matplotlib.pyplot as plt

        fig = plt.figure()
        fig.suptitle("Quantization noise vs Average Trace")
        ax = fig.add_subplot(2, 1, 1)
        ax.set_xlabel("Blocks")
        ax.set_yscale("log")
        perturbations_per_layer_id = list(self._perturbations.get_all().values())
        perturb = []
        max_bitwidths = []
        for perturbations_for_all_observed_qconfig_sequence_in_current_layer in perturbations_per_layer_id:
            qconfig_sequence = perturbations_for_all_observed_qconfig_sequence_in_current_layer.keys()
            max_bitwidth_qconfig = max(qconfig_sequence, key=lambda x: x.num_bits)
            perturb.append(perturbations_for_all_observed_qconfig_sequence_in_current_layer[max_bitwidth_qconfig])
            max_bitwidths.append(max_bitwidth_qconfig.num_bits)
        ax.plot(
            [
                (p / m / n).cpu().numpy()
                for p, m, n in zip(perturb, self._num_weights_per_layer, self._norm_weights_per_layer)
            ],
            label="normalized n-bit noise",
        )
        ax.plot([x.cpu().numpy() for x in perturb], label="n-bit noise")
        ax.plot(max_bitwidths, label="n")
        ax.plot(self._traces_per_layer.cpu().numpy(), label="trace")
        ax.plot([(n * p).cpu().numpy() for n, p in zip(self._traces_per_layer, perturb)], label="trace * noise")
        ax.legend()
        plt.savefig(os.path.join(self._dump_dir, "Quantization_noise_vs_Average_Trace"))

    def dump_bitwidth_graph(
        self,
        algo_ctrl: "QuantizationController",  # noqa: F821
        model: NNCFNetwork,
        groups_of_adjacent_quantizers: GroupsOfAdjacentQuantizers,
    ):
        from nncf.torch.quantization.precision_init.bitwidth_graph import BitwidthGraph

        bw_graph = BitwidthGraph(algo_ctrl, model, groups_of_adjacent_quantizers).get()
        nx_graph = add_adjust_padding_nodes(bw_graph, model)
        write_dot_graph(nx_graph, self._dump_dir / Path("bitwidth_graph.dot"))



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import OrderedDict

from nncf.common.hook_handle import add_op_to_registry


class InputType:
    INPUTS = "inputs"
    WEIGHTS = "weights"


class NNCFOperation:
    """
    The abstract class represents main building block for adding compression
    extensions to a model.
    """

    def __init__(self, name, trainable=True):
        """
        Initializes internal NNCF operation state

        :param name: unique operation name in algorithm scope.
        """
        self._call_pre_hooks = OrderedDict()
        self._name = name
        self._trainable = trainable

    @property
    def name(self):
        return self._name

    @property
    def trainable(self):
        return self._trainable

    @trainable.setter
    def trainable(self, value):
        self._trainable = value

    def build(self, input_shape, input_type, name, layer):
        """
        This method can be used to create weights that depend on the shape(s)
        of the input(s) and register them in the NNCF Wrapper `layer`. The method
        will be automatically called when NNCF Wrapper `layer` is built.

        :param input_shape: shape of the input
        :param input_type: type of the input identifies that inputs are layer weights
                           or inputs of the layer
        :param name: operation name
        :param layer: NNCF Wrapper layer, where the operation is registered
        :return: weights dictionary {weight name: weight value}
        """

    def call(self, inputs, weights, training):
        """
        The method performs the logic of applying the operation to the input tensors
        (which should be passed in as argument).

        :param inputs: input tensors
        :param weights: operation weights
        :param training: identifying that the model is training or evaluating
        :return: output tensors
        """
        raise NotImplementedError

    def register_hook_pre_call(self, hook):
        """
        Registers a hook which will be called before `call` function.
        The NNCF operation does not support serialization of the registered hooks.

        :param hook: callable object with following signatures:
                     hook(inputs) -> None or modified input
        :return: a handle that can be used to remove the hook form
                 the NNCF operation by calling handle.remove()
        """
        return add_op_to_registry(self._call_pre_hooks, hook)

    def __call__(self, *args, **kwargs):
        inputs = args[0]
        for hook in self._call_pre_hooks.values():
            result = hook(inputs)
            if result is not None:
                inputs = result
        return self.call(inputs, *args[1:], **kwargs)

    def get_config(self):
        return {"name": self._name}

    @classmethod
    def from_config(cls, config):
        return cls(**config)



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from collections import deque
from typing import Dict, List, Tuple

import openvino.runtime as ov
from openvino.runtime import opset13 as opset
from openvino.runtime.utils.node_factory import NodeFactory

from nncf.openvino.graph.model_transformer import OVModelTransformer
from nncf.openvino.graph.node_utils import get_parameter_node_name
from nncf.openvino.graph.node_utils import get_result_node_name


class OVModelBuilder:
    """
    The purpose of the ModelBuilder is to build a new OpenVINO model from input and output points.
    This Builder was created to reduce the number of model cloning that is required for ModelTransformer to work.
    """

    def __init__(self):
        self._node_factory = NodeFactory()

    @staticmethod
    def _create_parameter(node_name: str, node_input: ov.Input) -> ov.Node:
        """
        A method that contains steps to create a Parameter for a new model using a specific template.
        """
        port_id = node_input.get_index()
        parameter_name = get_parameter_node_name(node_name, port_id)
        return opset.parameter(
            shape=node_input.get_partial_shape(),
            dtype=node_input.get_element_type(),
            name=parameter_name,
        )

    @staticmethod
    def _create_result(node_name: str, node_output: ov.Input) -> ov.Node:
        """
        A method that contains steps to create a Result for a new model using a specific template.
        """
        port_id = node_output.get_index()
        result_name = get_result_node_name(node_name, port_id=port_id)
        result = opset.result(node_output, name=result_name)
        result.get_output_tensor(0).set_names({result_name})
        return result

    def _collect_graph_nodes(
        self,
        input_ids: List[Tuple[str, int]],
        output_ids: List[Tuple[str, int]],
        node_mapping: Dict[str, ov.Node],
    ) -> List[ov.Node]:
        """
        A method for aggregating layers to be further cloned.
        Aggregation is designed in such a way that layers are listed from right to left,
        as they pass from bottom to top. This is done in order to find all constants in the model and
        to start graph creation from them (as well as Parameter layers), because
        OpenVINO graph is created from top-down and cannot be created otherwise.

        Legend: w - weigths, c - convert, il/ih - input low/high, ol/oh - output low/high
        (w)
         |
        (c) (il) (ih) (ol) (oh)
          \   |   |   /   /
           (fake quantize) (parameter)
                  \           /
                  (convolution)
                        |
                    (result)
        Based on the above graph, the return value would look like this:
        [convolution, parameter, fake quantize, oh, ol, ih, il, c, w]

        :param input_ids: List of the points in the special format - (node_name, port_id).
            This helps to point to the precise part of the model that may be used to define the subgraph inputs.
        :param output_ids: List of the points in the special format - (node_name, port_id).
            This helps to point to the precise part of the model that may be used to define the subgraph outputs.
        :param node_mapping: Original nodes mapping.
        :return: List of the ov.Nodes to clone.
        """
        # Creating a list as a deque for FIFO layer acquisition and retrieval
        lookup_nodes = deque(node_mapping[n] for n, _ in output_ids)
        graph_nodes = []

        while lookup_nodes:
            lookup_node = lookup_nodes.popleft()
            lookup_name = lookup_node.get_friendly_name()
            node_inputs = lookup_node.inputs()
            graph_nodes.append(lookup_node)
            # Reversing to lookup nodes from right to left
            for node_input in reversed(node_inputs):
                port_id = node_input.get_index()
                if (lookup_name, port_id) in input_ids:
                    # We create Parameters here to avoid double creation in the future since it is not an original node,
                    # but we need to have it as input for next node.
                    parameter = self._create_parameter(lookup_name, node_input)
                    lookup_nodes.append(parameter)
                    continue
                parent_node = node_input.get_source_output().get_node()
                lookup_nodes.append(parent_node)

        return graph_nodes

    def build(
        self,
        input_ids: List[Tuple[str, int]],
        output_ids: List[Tuple[str, int]],
        node_mapping: Dict[str, ov.Node],
    ) -> ov.Model:
        """
        The basic method of the algorithm. This method uses an aggregated list of layers to be recreated.
        Let us take a graph of this kind as an example:

        Legend: w - weigths, c - convert, il/ih - input low/high, ol/oh - output low/high
        (w)
         |
        (c) (il) (ih) (ol) (oh)
          \   |   |   /   /
           (fake quantize) (parameter)
                  \           /
                  (convolution)
                        |
                    (result)

        The externally collected list of layers will look like this:
        [convolution, parameter, fake quantize, oh, ol, ih, il, c, w]

        Next, this list will be circled from right to left. At the same time, the list of already created layers
        will be filled from left to right, which will be used in the traversal step also, from left to right,
        in order to keep the order of the original layer inputs.
        For example:

            graph_nodes = [convolution, parameter, fake quantize, oh, ol, ih, il, c, w]
            clone_nodes = []

        *creating w - weight node.*
            graph_nodes = [convolution, parameter, fake quantize, oh, ol, ih, il, c]
            clone_nodes = [w]

        *creating c - convert node.
        Based on the .inputs() output, we'll use the already created w-weight node to fill in the convert input.
        As the result, weight node would be removed from the clone_nodes list and convert node would be placed here.*
            graph_nodes = [convolution, parameter, fake quantize, oh, ol, ih, il]
            clone_nodes = [c]

        *creating il/ih - input low/high, ol/oh - output low/high nodes.
        Since these nodes are constants and do not require any nodes as inputs, cloned nodes will not be used.*
            graph_nodes = [convolution, parameter, fake quantize, oh, ol, ih, il]
            clone_nodes = [c, il, ih, ol, oh]

        *creating fake quantize node.
        This node requires to have input values in a specific order.
        All previous nodes will be connected/used for fake quantize, from left to right.*
            graph_nodes = [convolution, parameter]
            clone_nodes = [f]

        *creating parameter node.
        In this step, the list of parameters will also be filled out with the new node.*
            graph_nodes = [convolution]
            clone_nodes = [f, parameter]

        *creating convolution node.
        This node also requires to have inputs in a specific order.
        All previous nodes will be connected/used for convolution, from left to right. Also,
        the outputs verification step will show here that one of the convolution outputs is in the output_ids list.
        This means that the Result node would be created and placed into the results list.*
            graph_nodes = []
            clone_nodes = [convolution]

        The last step is to create a subgraph model based on the parameters & results lists.

        :param input_ids: List of the points in the special format - (node_name, port_id).
            This helps to point to the precise part of the model that may be used to define the subgraph inputs.
        :param output_ids: List of the points in the special format - (node_name, port_id).
            This helps to point to the precise part of the model that may be used to define the subgraph outputs.
        :param node_mapping: Original nodes mapping.
        :return: Builded ov.Model based on parameters.
        """

        parameters, results = [], []
        clone_nodes = deque()

        # Collecting nodes that declares the graph.
        graph_nodes = self._collect_graph_nodes(input_ids, output_ids, node_mapping)

        while graph_nodes:
            graph_node = graph_nodes.pop()
            node_type = graph_node.get_type_name()
            node_name = graph_node.get_friendly_name()

            # To create the new OpenVINO nodes, we need to provide all possible layer attributes.
            attrs = graph_node.get_attributes()
            attrs["name"] = node_name

            if node_type == "Constant":
                # Constants creation is apart due to specific behavior.
                clone_node = OVModelTransformer._create_constant(
                    graph_node.get_data(), dtype=graph_node.get_element_type(), name=attrs["name"]
                )
            elif node_type == "Parameter":
                # We've created Parameter nodes on the previous step.
                clone_node = graph_node
                parameters.append(clone_node)
            else:
                # We have to have args as the inputs since all of them are nodes and are required to be as input.
                args = [clone_nodes.popleft() for _ in graph_node.inputs()]

                clone_node = self._node_factory.create(node_type, args, attrs)

                for node_output in clone_node.outputs():
                    port_id = node_output.get_index()
                    if (node_name, port_id) in output_ids:
                        result = self._create_result(node_name, node_output)
                        results.append(result)

            clone_nodes.append(clone_node)

        return ov.Model(results, parameters)



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import torch


def l1_filter_norm(weight_tensor, dim=0):
    """
    Calculates L1 for weight_tensor for the selected dimension.
    """
    weight_tensor = weight_tensor.transpose(0, dim).contiguous()
    return torch.norm(weight_tensor.view(weight_tensor.shape[0], -1), p=1, dim=1)


def l2_filter_norm(weight_tensor, dim=0):
    """
    Calculates L2 for weight_tensor for the selected dimension.
    """
    weight_tensor = weight_tensor.transpose(0, dim).contiguous()
    return torch.norm(weight_tensor.view(weight_tensor.shape[0], -1), p=2, dim=1)


def tensor_l2_normalizer(weight_tensor):
    norm = torch.sqrt(torch.sum(torch.abs(weight_tensor) ** 2))
    return weight_tensor / norm


def geometric_median_filter_norm(weight_tensor, dim=0):
    """
    Compute geometric median norm for filters.
    :param weight_tensor: tensor with weights
    :param dim: dimension of output channel
    :return: metric value for every weight from weights_tensor
    """
    weight_tensor = weight_tensor.transpose(0, dim).contiguous()
    filters_count = weight_tensor.size(0)
    weight_vec = weight_tensor.view(filters_count, -1)
    similarity_matrix = torch.cdist(weight_vec[None, :], weight_vec[None, :], p=2.0)
    return similarity_matrix.squeeze().sum(axis=0).to(weight_tensor.device)


FILTER_IMPORTANCE_FUNCTIONS = {
    "L2": l2_filter_norm,
    "L1": l1_filter_norm,
    "geometric_median": geometric_median_filter_norm,
}


def calculate_binary_mask(importance, threshold):
    return (importance >= threshold).float()



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import inspect
import types
from collections import defaultdict
from contextlib import contextmanager
from dataclasses import dataclass
from dataclasses import field
from itertools import chain
from types import MethodType
from types import TracebackType
from typing import Any, Callable, Dict, Iterator, List, Optional, Tuple, Type
from weakref import ReferenceType
from weakref import ref

import torch
from torch import Tensor
from torch import nn
from torch.overrides import TorchFunctionMode

from nncf.common.logging import nncf_logger as logger
from nncf.experimental.torch2.function_hook.handle_inner_functions import get_handle_inner_function
from nncf.experimental.torch2.function_hook.hook_storage import HookStorage
from nncf.experimental.torch2.function_hook.weak_map import WeakUnhashableKeyMap

IGNORED_FN_NAMES = [
    "__repr__",
    "_assert",
    "dim",
    "size",
    "is_floating_point",
    "_set_grad_enabled",
]


@dataclass
class OpMeta:
    """
    Metadata for an operation to be executed, including its name, the callable function,
    and any additional information.


    :param op_name: The name of the operation.
    :param func: The function to be executed for the operation.
    :param extra_info: A dictionary for storing any additional information about the operation.
    """

    op_name: str
    func: Callable[..., Any]
    extra_info: Dict[str, Any] = field(default_factory=lambda: dict())


def _get_full_fn_name(fn: Callable[..., Any]) -> str:
    """
    Get the full name of a function, including its module if applicable.

    :param fn: The function for which to get the full name.
    :return: The full name of the function.
    """
    if inspect.ismethoddescriptor(fn) or inspect.ismethod(fn):
        return fn.__qualname__
    if inspect.isbuiltin(fn) or inspect.isfunction(fn):
        return f"{fn.__module__}.{fn.__name__}"
    return f"{fn.__name__}"


def generate_normalized_op_name(module_name: str, fn_name: str, call_id: Optional[int] = None) -> str:
    """
    Returns a normalized name of operation.

    Returns: The string representation in the format "module_name/fn_name" or "module_name/fn_name/call_id".
    """
    if call_id is None:
        return f"{module_name}/{fn_name}"
    return f"{module_name}/{fn_name}/{call_id}"


class FunctionHookMode(TorchFunctionMode):
    """
    Executes pre- and post-hooks for PyTorch functions within a model's execution.

    This mode wraps the function calls in the model to allow custom hooks to be executed before
    and after the actual function calls.


    :param model: The PyTorch model to which the hooks will be applied.
    :param hook_storage: Storage for hooks to be executed.
    :param module_call_stack: A stack tracking the modules being called.
    :param nested_enter_count: A counter to track nested context manager entries.
    :param op_calls: A dictionary to track operation calls.
    """

    def __init__(self, model: nn.Module, hook_storage: HookStorage) -> None:
        """
        Initialize the FunctionHookMode.

        :param model: The PyTorch model to which the hooks will be applied.
        :param hook_storage: Storage for hooks to be executed.
        """
        super().__init__()
        self.hook_storage: HookStorage = hook_storage
        self.model: nn.Module = model
        self.module_call_stack: List[nn.Module] = []
        self.nested_enter_count = 0
        self.op_calls: Dict[str, int] = defaultdict(int)
        self.enabled: bool = True

        # Variables for hooks after constant nodes
        self.const_name_map: WeakUnhashableKeyMap[torch.Tensor, str] = WeakUnhashableKeyMap()
        for name, parameter in chain(self.model.named_parameters(), self.model.named_buffers()):
            self.const_name_map[parameter] = name
        self.in_process_const = False

        # Hook names
        self.hooks_module_to_group_name: Dict[ReferenceType[nn.Module], str] = {}
        self._get_named_hooks(self.hook_storage.pre_hooks, "pre_hook")
        self._get_named_hooks(self.hook_storage.post_hooks, "post_hook")

    def _get_named_hooks(self, storage: nn.ModuleDict, prefix: str) -> None:
        """
        Associates named hooks from the given module storage with a group name, updating
        the `hooks_module_to_group_name` dictionary for each hook.

        The group name is formatted as '{prefix}__{hook_key}[{hook_id}]', and any forward
        slashes in the hook key are replaced with dashes to avoid conflicts with separator
        characters used by the module hierarchy.

        :param storage: A module that stores hook modules.
        :param prefix: A string prefix used to group hooks by their context or origin.
        """
        for hook_key, hook_module_dict in storage.named_children():
            for hook_id, hook_module in hook_module_dict.named_children():
                # Replace / to avoid collision with module separator
                hook_name = hook_key.replace("/", "-")
                self.hooks_module_to_group_name[ref(hook_module)] = f"{prefix}__{hook_name}[{hook_id}]"

    def _get_wrapped_call(self, fn_call: MethodType) -> Callable[..., Any]:
        """
        Wrap a function call to include pushing to and popping from the module call stack.

        :param fn_call: The original function call to wrap.
        :return: The wrapped function call.
        """

        def wrapped_call(self_: nn.Module, *args: Tuple[Any, ...], **kwargs: Dict[str, Any]) -> Any:
            self.push_module_call_stack(self_)
            retval = fn_call.__func__(self_, *args, **kwargs)
            self.pop_module_call_stack()
            return retval

        return wrapped_call

    def __enter__(self) -> FunctionHookMode:
        """
        Enter the context manager.
        Wrapping the _call_impl function of each module on first nested enter.

        :return: The instance of FunctionHookMode.
        """
        super().__enter__()  # type: ignore
        if self.nested_enter_count == 0:
            # Wrap _call_impl function of instance each module.
            # Note: __call__ can`t not be overrided for instance, the function can be override only in class namespace.
            logger.debug("FunctionHookMode.__enter__: wrap _call_impl function")
            for _, module in self.model.named_modules():
                module._call_impl = types.MethodType(self._get_wrapped_call(module._call_impl), module)
            self.push_module_call_stack(self.model)
        self.nested_enter_count += 1
        return self

    def __exit__(
        self, exc_type: Optional[Type[BaseException]], exc_val: Optional[BaseException], exc_tb: Optional[TracebackType]
    ) -> None:
        """
        Exit the context manager, unwrapping the _call_impl function of each module.

        :param exc_type: Exception type.
        :param exc_val: Exception value.
        :param exc_tb: Traceback.
        """
        self.nested_enter_count -= 1
        if self.nested_enter_count == 0:
            # Unwrap _call_impl functions
            logger.debug("FunctionHookMode.__exit__: unwrap _call_impl function")
            for _, module in self.model.named_modules():
                module.__dict__.pop("_call_impl")
            self.pop_module_call_stack()
        super().__exit__(exc_type, exc_val, exc_tb)  # type: ignore

    def __torch_function__(
        self,
        func: Callable[..., Any],
        types: List[Type[Any]],
        args: Tuple[Any, ...] = (),
        kwargs: Optional[Dict[str, Any]] = None,
    ) -> Any:
        """
        Override the __torch_function__ method to add pre- and post-hook execution.

        :param func: The function being called.
        :param types: List of types.
        :param args: The arguments to the function.
        :param kwargs: The keyword arguments to the function.
        :return: The output of the function call after hooks have been executed.
        """
        kwargs = kwargs or {}

        fn_name = func.__name__

        # WA: to catch nested calls for some functions
        # https://github.com/pytorch/pytorch/issues/55093
        fn_for_nested_call = get_handle_inner_function(func)
        if fn_for_nested_call is not None:
            with self:
                return fn_for_nested_call(*args, **kwargs)

        if not self.enabled or fn_name in IGNORED_FN_NAMES:
            return func(*args, **kwargs)

        op_name = self.get_current_executed_op_name(fn_name)
        full_fn_name = _get_full_fn_name(func)
        logger.debug(f"FunctionHookMode.__torch_function__: {full_fn_name=} {op_name=}")
        self.register_op(fn_name)
        op_meta = OpMeta(op_name=op_name, func=func)
        args, kwargs = self.execute_pre_hooks(args, kwargs, op_meta)
        output = func(*args, **kwargs)
        output = self.execute_post_hooks(output, op_meta)
        return output

    def push_module_call_stack(self, module: nn.Module) -> None:
        """
        Push a module to the call stack and initialize its operation calls.

        :param module: The module to push onto the call stack.
        """
        assert isinstance(module, nn.Module)
        self.module_call_stack.append(module)
        module_name = self.get_current_relative_name()
        logger.debug(f"FunctionHookMode.push_module_call_stack: {module_name=}")

    def pop_module_call_stack(self) -> None:
        """
        Pop a module from the call stack and remove its operation calls.
        """
        module_name = self.get_current_relative_name()
        self.module_call_stack.pop()
        logger.debug(f"FunctionHookMode.pop_module_call_stack: {module_name=}")

    def get_current_relative_name(self) -> str:
        """
        Get the name of the current module being executed.

        :return: The name of the current module.
        """
        relative_module_names = []
        prev_module = self.module_call_stack[0]

        for module in self.module_call_stack[1:]:
            hook_name = self.hooks_module_to_group_name.get(ref(module))
            if hook_name is not None:
                relative_module_names.append(hook_name)
            else:
                for n, m in prev_module.named_children():
                    if m is module:
                        relative_module_names.append(n)
                        break
            prev_module = module

        return "/".join(relative_module_names)

    def get_current_executed_op_name(self, fn_name: str) -> str:
        """
        Get the name of the current operation being executed.

        :param fn_name: The function name of the operation.
        :return: The name of the operation.
        """
        module_name = self.get_current_relative_name()
        op_name = generate_normalized_op_name(module_name, fn_name)
        call_id = self.op_calls[op_name]
        return generate_normalized_op_name(module_name, fn_name, call_id)

    def register_op(self, fn_name: str) -> None:
        """
        Register an operation call for the current module and increment call counter.

        :param fn_name: The function name of the operation.
        """
        module_name = self.get_current_relative_name()
        op_name = generate_normalized_op_name(module_name, fn_name)
        self.op_calls[op_name] += 1

    def execute_hooks_for_parameter(self, value: torch.Tensor) -> torch.Tensor:
        """
        Executes post-hooks for a model parameter if a hook is defined for it.
        If the input is not a `torch.nn.Parameter`, or if no hook is defined, the original tensor is returned unchanged.

        :param value: The tensor to which the post-hook will be applied..
        :return: The processed tensor with the applied post-hook, if applicable.
        """
        if not isinstance(value, torch.nn.Parameter):
            return value

        name_in_model = self.const_name_map.get(value, None)
        if name_in_model is not None and not self.in_process_const:
            self.in_process_const = True
            value = self.hook_storage.execute_post_function_hooks(name_in_model.replace(".", ":"), 0, value)
            self.in_process_const = False
        return value

    def process_parameters(self, args: List[Any], kwargs: Dict[str, Any]) -> Tuple[List[Any], Dict[str, Any]]:
        """
        Execute post-hooks for parameters.

        :param args: The arguments to the function.
        :param kwargs: The keyword arguments to the function.
        :return: The modified arguments and keyword arguments after pre-hooks.
        """
        for idx, value in enumerate(args):
            args[idx] = self.execute_hooks_for_parameter(value)
        for kw_name, value in kwargs.items():
            kwargs[kw_name] = self.execute_hooks_for_parameter(value)
        return args, kwargs

    def execute_pre_hooks(
        self, args: Tuple[Any, ...], kwargs: Dict[str, Any], op_meta: OpMeta
    ) -> Tuple[Tuple[Any, ...], Dict[str, Any]]:
        """
        Execute pre-hooks for the operation.

        :param args: The arguments to the function.
        :param kwargs: The keyword arguments to the function.
        :param op_meta: Metadata for the operation.
        :return: The modified arguments and keyword arguments after pre-hooks.
        """
        _args: List[Any] = list(args)
        with self:
            _args, kwargs = self.process_parameters(_args, kwargs)

            for idx, value in enumerate(_args):
                _args[idx] = self.hook_storage.execute_pre_function_hooks(op_meta.op_name, idx, value)

            for port_id, kw_name in enumerate(kwargs, start=len(_args)):
                kwargs[kw_name] = self.hook_storage.execute_pre_function_hooks(
                    op_meta.op_name, port_id, kwargs[kw_name]
                )
        return tuple(_args), kwargs

    def process_post_function_hooks_for_value(self, value: Any, op_meta: OpMeta, port_id: int) -> Any:
        """
        Executes post-hooks on a value after an operation.

        :param value: The output value.
        :param op_meta: Metadata about the operation.
        :param port_id: The id of output port.
        :return: The modified output after post-hooks.
        """
        return self.hook_storage.execute_post_function_hooks(op_meta.op_name, port_id, value)

    def execute_post_hooks(self, output: Any, op_meta: OpMeta) -> Any:
        """
        Execute post-hooks for the operation.

        :param output: The output of the function.
        :param op_meta: Metadata for the operation.
        :return: The modified output after post-hooks.
        """
        with self:
            cls_tuple = None
            if isinstance(output, tuple):
                # Need to return named tuples like torch.return_types.max
                cls_tuple = type(output)
                output = list(output)

            if isinstance(output, list):
                for idx, value in enumerate(output):
                    output[idx] = self.process_post_function_hooks_for_value(value, op_meta, idx)
                if cls_tuple is not None:
                    output = cls_tuple(output)
            else:
                output = self.process_post_function_hooks_for_value(output, op_meta, 0)
        return output

    def process_model_inputs(self, args: Tuple[Any], kwargs: Dict[str, Any]) -> Tuple[Tuple[Any], Dict[str, Any]]:
        """
        Processes the input arguments for the model's forward function.

        :param args: Positional arguments passed to the model's forward method.
        :param kwargs: Keyword arguments passed to the model's forward method.
        :return: The processed arguments, with hooks applied to any tensor inputs.
        """
        forward_signature = inspect.signature(self.model.forward)
        bound_arguments = forward_signature.bind(*args, **kwargs)

        # Hooks available only for named arguments
        for name, value in bound_arguments.arguments.items():
            if isinstance(value, Tensor):
                bound_arguments.arguments[name] = self.execute_hooks_for_model_input(name, value)

        return bound_arguments.args, bound_arguments.kwargs

    def execute_hooks_for_model_input(self, name: str, value: Any) -> Any:
        """
        Executes the post-hooks for an input tensor.

        :param name: The name of the input argument.
        :param value: The value of the input argument.
        :return: The processed value after the hook is executed.
        """
        return self.hook_storage.execute_post_function_hooks(name, 0, value)

    def process_model_outputs(self, outputs: Any) -> Any:
        """
        Processes the outputs from the model, applying pre-hooks to any tensors found in the output.

        :param outputs: The outputs returned by the model's forward method.
        :return: The processed outputs with hooks applied.
        """
        if isinstance(outputs, Tensor):
            return self.execute_hooks_for_model_output("output", outputs)

        cls_tuple = None
        if isinstance(outputs, tuple):
            cls_tuple = type(outputs)
            outputs = list(outputs)
        if isinstance(outputs, list):
            outputs = list(outputs)
            for idx, val in enumerate(outputs):
                if isinstance(val, Tensor):
                    outputs[idx] = self.execute_hooks_for_model_output(f"output_{idx}", val)
        if cls_tuple is not None:
            outputs = cls_tuple(outputs)
        return outputs

    def execute_hooks_for_model_output(self, name: str, value: Any) -> Any:
        """
        Executes the pre-hooks for an input tensor.

        :param name: The name of the input argument.
        :param value: The value of the input argument.
        :return: The processed value after the hook is executed.
        """
        return self.hook_storage.execute_pre_function_hooks(name, 0, value)

    @contextmanager
    def disable(self) -> Iterator[None]:
        """
        Temporarily disables the function tracing and execution hooks within a context.

        :return: A context manager which temporary disables the function tracing and execution hooks within a context.
        """
        ret = self.enabled
        self.enabled = False
        yield
        self.enabled = ret



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import abstractmethod
from typing import List

import numpy as np

from nncf.common.collector import StatisticsCollector
from nncf.common.sparsity.statistics import SparsifiedLayerSummary
from nncf.common.sparsity.statistics import SparsifiedModelStatistics


class WeightDescription:
    """
    Contains information about the weight of the model.
    """

    def __init__(self, name: str, shape: List[int], num_nonzero: int, is_sparse: bool):
        """
        Initializes the description of the weight.

        :param name: Identifier of the weight.
        :param shape: Shape of the weight.
        :param num_nonzero: Number of nonzero parameters of the weight.
        :param is_sparse: It is a bool value that specifies if the
            weight is used in the sparsity algorithm or not.
        """
        self._name = name
        self._shape = shape
        self._num_nonzero = num_nonzero
        self._is_sparse = is_sparse

        self._num_params = np.prod(shape).item()
        self._num_zero = self._num_params - self._num_nonzero
        self._sparsity_level = self._num_zero / self._num_params

    @property
    def name(self) -> str:
        return self._name

    @property
    def shape(self) -> List[int]:
        return self._shape

    @property
    def num_params(self) -> int:
        return self._num_params

    @property
    def num_nonzero(self) -> int:
        return self._num_nonzero

    @property
    def num_zero(self) -> int:
        return self._num_zero

    @property
    def sparsity_level(self) -> float:
        return self._sparsity_level

    @property
    def is_sparse(self) -> bool:
        return self._is_sparse


def _calculate_sparsity_level_for_model(
    weight_descriptions: List[WeightDescription],
) -> float:
    """
    Calculates the sparsity level for the whole model.

    :param weight_descriptions: Descriptions for weights of the model.
    :return: Sparsity level for the whole model.
    """
    total_params = sum(w.num_params for w in weight_descriptions)
    total_num_zero = sum(w.num_zero for w in weight_descriptions)
    sparsity_level = total_num_zero / total_params

    return sparsity_level


class BaseSparseModelStatisticsCollector(StatisticsCollector):
    """
    Base class for the sparse model statistics collector.
    """

    @abstractmethod
    def _collect_weights_descriptions(self) -> List[WeightDescription]:
        """
        Collects descriptions of the weights of the model.

        :return: Descriptions of the weights of the model.
        """

    def collect(self) -> SparsifiedModelStatistics:
        """
        Collects statistics for the sparse model.

        :return: An instance of the `SparsifiedModelStatistics` class.
        """
        weights_descriptions = self._collect_weights_descriptions()
        sparsity_level_for_model = _calculate_sparsity_level_for_model(weights_descriptions)

        total_params = sum(w.num_params for w in weights_descriptions if w.is_sparse)
        total_num_zero = sum(w.num_zero for w in weights_descriptions if w.is_sparse)
        sparsity_level_for_sparse_layers = total_num_zero / total_params

        sparse_layers_summary = []
        for w in weights_descriptions:
            if not w.is_sparse:
                continue

            weight_percentage = 100 * (w.num_params / total_params)
            sparse_layers_summary.append(SparsifiedLayerSummary(w.name, w.shape, w.sparsity_level, weight_percentage))

        sparse_model_stats = SparsifiedModelStatistics(
            sparsity_level_for_model,
            sparsity_level_for_sparse_layers,
            sparse_layers_summary,
        )

        return sparse_model_stats



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Optional, Tuple

import numpy as np
import openvino.runtime as ov

from nncf.common.graph.transformations.commands import Command
from nncf.common.graph.transformations.commands import TargetPoint
from nncf.common.graph.transformations.commands import TargetType
from nncf.common.graph.transformations.commands import TransformationCommand
from nncf.common.graph.transformations.commands import TransformationType
from nncf.openvino.graph.node_utils import InplaceInsertionFnType
from nncf.quantization.fake_quantize import FakeConvertParameters
from nncf.quantization.fake_quantize import FakeQuantizeParameters


class OVTargetPoint(TargetPoint):
    def __init__(self, target_type: TargetType, target_node_name: str, port_id: int):
        super().__init__(target_type)
        self.target_node_name = target_node_name
        self.port_id = port_id

    def __eq__(self, other: "OVTargetPoint") -> bool:
        return (
            isinstance(other, OVTargetPoint)
            and self.type == other.type
            and self.target_node_name == other.target_node_name
            and self.port_id == other.port_id
        )

    def __hash__(self) -> int:
        return hash((self.target_node_name, self.port_id, self._target_type))


class OVInsertionCommand(TransformationCommand):
    def __init__(self, target_point: OVTargetPoint):
        super().__init__(TransformationType.INSERT, target_point)


class OVOutputInsertionCommand(OVInsertionCommand):
    def __init__(self, target_point: OVTargetPoint, output_dtype: ov.Type = ov.Type.f32):
        super().__init__(target_point)
        self.output_dtype = output_dtype

    def union(self, other: "TransformationCommand") -> "TransformationCommand":
        # Have a look at nncf/torch/graph/transformations/commands/PTInsertionCommand
        raise NotImplementedError()


class OVInplaceFnInsertionCommand(OVInsertionCommand):
    def __init__(
        self,
        target_point: OVTargetPoint,
        inplace_op_fn: InplaceInsertionFnType,
        fn_output_port_id: int,
        last_inplace_node_name: str,
        output_dtype: Optional[ov.Type] = None,
    ):
        super().__init__(target_point)
        self.inplace_op_fn = inplace_op_fn
        self.fn_output_port_id = fn_output_port_id
        self.last_inplace_node_name = last_inplace_node_name
        self.output_dtype = output_dtype

    def union(self, other: "TransformationCommand") -> "TransformationCommand":
        # Have a look at nncf/torch/graph/transformations/commands/PTInsertionCommand
        raise NotImplementedError()


class OVFQNodeRemovingCommand(TransformationCommand):
    """
    Removes FakeQuantize nodes from the model.
    """

    def __init__(self, target_point: OVTargetPoint):
        """
        :param target_point: The TargetPoint instance for the layer that contains information for removing.
        """
        super().__init__(TransformationType.REMOVE, target_point)


class OVQuantizerInsertionCommand(OVInsertionCommand):
    def __init__(self, target_point: OVTargetPoint, quantizer_parameters: FakeQuantizeParameters):
        super().__init__(target_point)
        self.quantizer_parameters = quantizer_parameters


class OVConvertInsertionCommand(OVInsertionCommand):
    def __init__(self, target_point: OVTargetPoint, convert_parameters: FakeConvertParameters):
        super().__init__(target_point)
        self.convert_parameters = convert_parameters


class OVBiasCorrectionCommand(TransformationCommand):
    """
    Corrects bias value in the model based on the input value.
    """

    def __init__(self, target_point: OVTargetPoint, bias_value: np.ndarray):
        """
        :param target_point: The TargetPoint instance for the correction that contains layer's information.
        :param bias_value: The bias shift value (numpy format) that will be added to the original bias value.
        """
        super().__init__(TransformationType.CHANGE, target_point)
        self.bias_value = bias_value


class OVWeightUpdateCommand(TransformationCommand):
    """
    Updates weight value in the model.
    """

    def __init__(self, target_point: OVTargetPoint, weight_value: np.ndarray):
        """
        :param target_point: Target point.
        :param weight_value: New weight value.
        """
        super().__init__(TransformationType.CHANGE, target_point)
        self.weight_value = weight_value


class OVModelExtractionCommand(Command):
    """
    Extracts sub-graph based on the sub-model input and output names.
    """

    def __init__(self, input_ids: List[Tuple[str, int]], output_ids: List[Tuple[str, int]]):
        """
        :param input_ids: List of the input IDs: pairs of node names and correspondent input port ids.
            Each pair denotes the sub-graph beginning.
        :param output_ids: List of the output IDs: pairs of node names and correspondent output port ids.
            Each pair denotes the sub-graph ending.
        """
        super().__init__(TransformationType.EXTRACT)
        self.input_ids = input_ids
        self.output_ids = output_ids


class OVStateLessModelExtractionCommand(Command):
    """
    Extracts stateless sub-graph based on the sub-model input and output names.
    """

    def __init__(self, input_ids: List[Tuple[str, int]], output_ids: List[Tuple[str, int]]):
        """
        :param input_ids: List of the input IDs: pairs of node names and correspondent output port ids.
            Each pair denotes the sub-graph beginning.
        :param output_ids: List of the output IDs: pairs of node names and correspondent output port ids.
            Each pair denotes the sub-graph ending.
        """
        super().__init__(TransformationType.EXTRACT)
        self.input_ids = input_ids
        self.output_ids = output_ids


class OVBiasInsertionCommand(TransformationCommand):
    """
    Inserts bias for the corresponding node.
    """

    def __init__(self, target_point: OVTargetPoint, bias_value: np.ndarray):
        """
        :param target_point: The TargetPoint instance for the insertion that contains layer's information.
        :param bias_value: Constant value for the bias layer.
        """
        super().__init__(TransformationType.INSERT, target_point)
        self.bias_value = bias_value


class OVMultiplyInsertionCommand(OVInsertionCommand):
    """
    Inserts Multiply nodes before the corresponding nodes.
    """

    def __init__(
        self,
        target_point: OVTargetPoint,
        scale_value: np.ndarray,
        destination_node_names: List[str],
        multiply_node_name: str,
    ):
        """
        :param target_point: The TargetPoint instance for the insertion that contains layer's information.
        :param scale_value: Scale value for Multiply layer.
        :param destination_node_names: New layer consumers.
        :param multiply_node_name: New layer name.
        """
        super().__init__(target_point)
        self.scale_value = scale_value
        self.destination_node_names = destination_node_names
        self.multiply_node_name = multiply_node_name


class OVUpdateIfBodyCommand(TransformationCommand):
    """
    Updates If node body.
    """

    def __init__(self, target_point: OVTargetPoint, body_model: ov.Model):
        """
        :param target_point: The TargetPoint instance for the change that contains layer's information.
        :param body_model: A new model to set.
        """
        super().__init__(TransformationType.CHANGE, target_point)
        self.subgraph_model = body_model


class OVExtractIfBodyCommand(Command):
    """
    Extracts If node body.
    """

    def __init__(self, if_node_name: str, if_body_condition: bool):
        """
        :param target_point: The TargetPoint instance for the extraction that contains layer's information.
        :param if_body_condition: If true extracts then body, else - else body.
        """
        super().__init__(TransformationType.EXTRACT)
        self.if_node_name = if_node_name
        self.if_body_condition = if_body_condition



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import ctypes
import importlib
import inspect
from collections import OrderedDict
from functools import partial
from functools import partialmethod
from typing import Any, Callable, Dict, List, Optional, Tuple, Union


class Patcher:
    """
    Method patch helper. Implements monkey patching approach.
    Note: imported from OTX.
    """

    def __init__(self) -> None:
        self._patched: Dict[Tuple[Any, str], List[Tuple[Any, Callable[..., Any]]]] = OrderedDict()

    def patch(  # noqa: C901
        self,
        obj_cls: Union[Any],
        wrapper: Callable[..., Any],
        *,
        force: bool = True,
    ) -> None:
        """
        Apply patching
        :param obj_cls: Function to be overriden.
        :param wrapper: Wrapper function to override with.
        :param force: Whether to override previously applied patches or not.
        """

        obj_cls, fn_name = self.import_obj(obj_cls)

        # wrap only if function does exist
        n_args = len(inspect.getfullargspec(obj_cls.__getattribute__)[0])
        if n_args == 1:
            try:
                fn = obj_cls.__getattribute__(fn_name)
            except AttributeError:
                return
            self._patch_module_fn(obj_cls, fn_name, fn, wrapper, force)
        else:
            if inspect.isclass(obj_cls):
                try:
                    fn = obj_cls.__getattribute__(obj_cls, fn_name)  # type: ignore
                except AttributeError:
                    return
                self._patch_class_fn(obj_cls, fn_name, fn, wrapper, force)
            else:
                try:
                    fn = obj_cls.__getattribute__(fn_name)
                except AttributeError:
                    return
                self._patch_instance_fn(obj_cls, fn_name, fn, wrapper, force)

    def unpatch(self, obj_cls: Any = None, depth: int = 0) -> None:
        """
        Undo patching
        :param obj_cls: Function to be unpatched.
        :param depth: How many patches to undo, depth=0 to undo all of them
        """

        def _unpatch(obj: Any, fn_name: str, key: Tuple[Any, str], depth: int) -> None:
            if depth == 0:
                depth = len(self._patched[key])
            keep = len(self._patched[key]) - depth
            origin_fn = self._patched[key].pop(-depth)[0]
            while self._patched[key] and len(self._patched[key]) > keep:
                self._patched[key].pop()
            if not self._patched[key]:
                self._patched.pop(key)

            if isinstance(obj, int):
                obj = ctypes.cast(obj, ctypes.py_object).value
            setattr(obj, fn_name, origin_fn)

        if obj_cls is not None:
            obj_cls, fn_name = self.import_obj(obj_cls)
            n_args = len(inspect.getfullargspec(obj_cls.__getattribute__)[0])
            if n_args == 1:
                key = (obj_cls.__name__, fn_name)
            else:
                if inspect.isclass(obj_cls):
                    obj_cls_path = obj_cls.__module__ + "." + obj_cls.__name__
                    key = (obj_cls_path, fn_name)
                else:
                    key = (id(obj_cls), fn_name)
            _unpatch(obj_cls, fn_name, key, depth)
            return

        for key in list(self._patched.keys()):
            obj, fn_name = key
            if isinstance(obj, int):
                obj = ctypes.cast(obj, ctypes.py_object).value
            else:
                obj, fn_name = self.import_obj(".".join([obj, fn_name]))
            _unpatch(obj, fn_name, key, depth)

    def import_obj(self, obj_cls: Union[str, Any]) -> Tuple[Any, str]:  # noqa: C901
        """Object import helper."""
        if isinstance(obj_cls, str):
            fn_name = obj_cls.split(".")[-1]
            obj_cls = ".".join(obj_cls.split(".")[:-1])
        else:
            if "_partialmethod" in obj_cls.__dict__:
                while "_partialmethod" in obj_cls.__dict__:
                    obj_cls = obj_cls._partialmethod.keywords["__fn"]
            while isinstance(obj_cls, (partial, partialmethod)):
                obj_cls = obj_cls.keywords["__fn"]

            if inspect.ismodule(obj_cls):
                fn = obj_cls.keywords["__fn"]
                obj_cls = fn = obj_cls.keywords["__obj_cls"]
                fn_name = fn.__name__
            elif inspect.ismethod(obj_cls):
                fn_name = obj_cls.__name__
                obj_cls = obj_cls.__self__
            elif isinstance(obj_cls, (staticmethod, classmethod)):
                obj_cls = obj_cls.__func__
                fn_name = obj_cls.__name__
                obj_cls = ".".join([obj_cls.__module__] + obj_cls.__qualname__.split(".")[:-1])
            else:
                fn_name = obj_cls.__name__
                obj_cls = ".".join([obj_cls.__module__] + obj_cls.__qualname__.split(".")[:-1])

        if isinstance(obj_cls, str):
            try:
                obj_cls = importlib.import_module(obj_cls)
            except ModuleNotFoundError:
                module = ".".join(obj_cls.split(".")[:-1])
                obj_cls = obj_cls.split(".")[-1]
                obj_cls = getattr(importlib.import_module(module), obj_cls)
        return obj_cls, fn_name

    def _patch_module_fn(self, obj_cls: Any, fn_name: str, fn: Any, wrapper: Callable[..., Any], force: bool) -> None:
        def helper(*args, **kwargs):  # type: ignore
            obj_cls = kwargs.pop("__obj_cls")
            fn = kwargs.pop("__fn")
            wrapper = kwargs.pop("__wrapper")
            return wrapper(obj_cls, fn, *args, **kwargs)

        assert len(inspect.getfullargspec(obj_cls.__getattribute__)[0]) == 1
        obj_cls_path = obj_cls.__name__
        key = (obj_cls_path, fn_name)
        fn_ = self._initialize(key, force)
        if fn_ is not None:
            fn = fn_
        setattr(obj_cls, fn_name, partial(helper, __wrapper=wrapper, __fn=fn, __obj_cls=obj_cls))
        self._patched[key].append((fn, wrapper))

    def _patch_class_fn(self, obj_cls: Any, fn_name: str, fn: Any, wrapper: Callable[..., Any], force: bool) -> None:
        if isinstance(fn, (staticmethod, classmethod)):

            def helper(*args, **kwargs):  # type: ignore
                wrapper = kwargs.pop("__wrapper")
                fn = kwargs.pop("__fn")
                obj_cls = kwargs.pop("__obj_cls")
                if isinstance(args[0], obj_cls):
                    return wrapper(args[0], fn.__get__(args[0]), *args[1:], **kwargs)
                return wrapper(obj_cls, fn.__get__(obj_cls), *args, **kwargs)

        elif isinstance(fn, type(all.__call__)):  # type: ignore

            def helper(self, *args, **kwargs):  # type: ignore
                kwargs.pop("__obj_cls")
                wrapper = kwargs.pop("__wrapper")
                fn = kwargs.pop("__fn")
                return wrapper(self, fn, *args, **kwargs)

        else:

            def helper(self, *args, **kwargs):  # type: ignore
                kwargs.pop("__obj_cls")
                wrapper = kwargs.pop("__wrapper")
                fn = kwargs.pop("__fn")
                return wrapper(self, fn.__get__(self), *args, **kwargs)

        assert len(inspect.getfullargspec(obj_cls.__getattribute__)[0]) == 2
        obj_cls_path = obj_cls.__module__ + "." + obj_cls.__name__
        key = (obj_cls_path, fn_name)
        fn_ = self._initialize(key, force)
        if fn_ is not None:
            fn = fn_
        setattr(
            obj_cls,
            fn_name,
            partialmethod(helper, __wrapper=wrapper, __fn=fn, __obj_cls=obj_cls),
        )
        self._patched[key].append((fn, wrapper))

    def _patch_instance_fn(self, obj_cls: Any, fn_name: str, fn: Any, wrapper: Callable[..., Any], force: bool) -> None:
        def helper(ctx, *args, **kwargs):  # type: ignore
            fn = kwargs.pop("__fn")
            wrapper = kwargs.pop("__wrapper")
            return wrapper(ctx, fn, *args, **kwargs)

        assert len(inspect.getfullargspec(obj_cls.__getattribute__)[0]) == 2
        obj_cls_path = id(obj_cls)
        key = (obj_cls_path, fn_name)
        fn_ = self._initialize(key, force)
        if fn_ is not None:
            fn = fn_
        setattr(obj_cls, fn_name, partialmethod(helper, __wrapper=wrapper, __fn=fn).__get__(obj_cls))
        self._patched[key].append((fn, wrapper))

    def _initialize(self, key: Tuple[int, str], force: bool) -> Optional[Any]:
        fn = None
        if key not in self._patched:
            self._patched[key] = []
        if force:
            while self._patched[key]:
                fn, *_ = self._patched[key].pop()
        return fn


PATCHER = Patcher()



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import tensorflow as tf

from nncf.common.compression import BaseCompressionAlgorithmController
from nncf.common.sparsity.controller import SparsityController
from nncf.tensorflow.graph.metatypes import keras_layers as layer_metatypes
from nncf.tensorflow.sparsity.utils import strip_model_from_masks

SPARSITY_LAYER_METATYPES = [
    layer_metatypes.TFConv1DLayerMetatype,
    layer_metatypes.TFConv2DLayerMetatype,
    layer_metatypes.TFConv3DLayerMetatype,
    layer_metatypes.TFDepthwiseConv1DSubLayerMetatype,
    layer_metatypes.TFDepthwiseConv2DSubLayerMetatype,
    layer_metatypes.TFDepthwiseConv3DSubLayerMetatype,
    layer_metatypes.TFDepthwiseConv2DLayerMetatype,
    layer_metatypes.TFConv1DTransposeLayerMetatype,
    layer_metatypes.TFConv2DTransposeLayerMetatype,
    layer_metatypes.TFConv3DTransposeLayerMetatype,
    layer_metatypes.TFSeparableConv1DLayerMetatype,
    layer_metatypes.TFSeparableConv2DLayerMetatype,
    layer_metatypes.TFEmbeddingLayerMetatype,
    layer_metatypes.TFLocallyConnected1DLayerMetatype,
    layer_metatypes.TFLocallyConnected2DLayerMetatype,
    layer_metatypes.TFDenseLayerMetatype,
]


class BaseSparsityController(BaseCompressionAlgorithmController, SparsityController):
    """
    Serves as a handle to the additional modules, parameters and hooks inserted
    into the original uncompressed model to enable sparsity-specific compression.
    Hosts entities that are to be used during the training process, such as
    compression scheduler and compression loss.
    """

    def __init__(self, target_model, op_names):
        super().__init__(target_model)
        self._op_names = op_names

    def strip_model(self, model: tf.keras.Model, do_copy: bool = False) -> tf.keras.Model:
        # Transform model for sparsity creates copy of the model.
        return strip_model_from_masks(model, self._op_names)



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any, Dict, List

import torch
from torch import nn

from nncf.torch.layer_utils import COMPRESSION_MODULES
from nncf.torch.layer_utils import StatefullModuleInterface
from nncf.torch.sparsity.functions import apply_binary_mask as apply_binary_mask_impl
from nncf.torch.utils import is_tracing_state


@COMPRESSION_MODULES.register()
class BinaryMask(nn.Module, StatefullModuleInterface):
    SHAPE_KEY = "shape"

    def __init__(self, shape: List[int]):
        super().__init__()
        self.register_buffer("_binary_mask", torch.ones(shape))
        self.frozen = False

    @property
    def binary_mask(self) -> torch.Tensor:
        return self._binary_mask

    @binary_mask.setter
    def binary_mask(self, tensor: torch.Tensor):
        with torch.no_grad():
            self._binary_mask.set_(tensor)

    def forward(self, weight):
        if is_tracing_state():
            return weight.mul(self.binary_mask)
        tmp_tensor = self._calc_training_binary_mask(weight)
        return apply_binary_mask_impl(tmp_tensor, weight)

    def _calc_training_binary_mask(self, weight):
        return self.binary_mask

    def apply_binary_mask(self, weight):
        return apply_binary_mask_impl(self.binary_mask, weight)

    def get_config(self) -> Dict[str, Any]:
        return {self.SHAPE_KEY: list(self.binary_mask.shape)}

    @classmethod
    def from_config(cls, state: Dict[str, Any]) -> "BinaryMask":
        return BinaryMask(state[cls.SHAPE_KEY])



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Make these names available from nncf.common.graph directly
from nncf.common.graph.graph import LayerName as LayerName
from nncf.common.graph.graph import NNCFGraph as NNCFGraph
from nncf.common.graph.graph import NNCFGraphEdge as NNCFGraphEdge
from nncf.common.graph.graph import NNCFGraphPatternIO as NNCFGraphPatternIO
from nncf.common.graph.graph import NNCFNode as NNCFNode
from nncf.common.graph.graph import NNCFNodeName as NNCFNodeName



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, TypeVar

import nncf

TensorType = TypeVar("TensorType")
DeviceType = TypeVar("DeviceType")
TensorElementsType = TypeVar("TensorElementsType")


class NNCFTensor:
    """
    An interface of framework specific tensors for common NNCF algorithms.
    """

    def __init__(self, tensor: TensorType):
        self._tensor = tensor

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, NNCFTensor):
            raise nncf.InternalError("Attempt to compare NNCFTensor with a non-NNCFTensor object")
        return self._tensor == other.tensor

    @property
    def tensor(self) -> TensorType:  # type: ignore
        return self._tensor

    @property
    def shape(self) -> List[int]:
        if self._tensor is None:
            raise nncf.InternalError("Attempt to get shape of empty NNCFTensor")
        return self._tensor.shape  # type: ignore

    @property
    def device(self) -> DeviceType:  # type: ignore
        raise NotImplementedError

    def is_empty(self) -> bool:
        raise NotImplementedError



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import partial
from typing import List, Set, Tuple, Type, Union

from nncf.common.graph import NNCFGraph
from nncf.common.graph import NNCFNode
from nncf.common.graph.operator_metatypes import OperatorMetatype
from nncf.common.logging import nncf_logger
from nncf.common.pruning.utils import traverse_function


def get_concat_axis(input_shapes: List[List[int]], output_shapes: List[List[int]]) -> int:
    """
    Returns concatenation axis by given input and output shape of concat node.

    :param input_shapes: Input_shapes of given concat node.
    :param output_shapes: Output_shapes of given concat node.
    :returns: Concatenation axis of given concat node.
    """
    axis = None
    none_dim = None
    for idx, (dim_in, dim_out) in enumerate(zip(input_shapes[0], output_shapes[0])):
        if dim_in != dim_out:
            axis = idx
            break
        if dim_in is None:
            none_dim = idx

    if axis is None:
        if none_dim is None:
            axis = -1
            nncf_logger.debug("Identity concat node detected")
        else:
            axis = none_dim

    return axis


def get_first_nodes_of_type(graph: NNCFGraph, op_types: List[str]) -> List[NNCFNode]:
    """
    Looking for first node in graph with type in `op_types`.
    First == layer with type in `op_types`, that there is a path from the input such that there are no other
    operations with type in `op_types` on it.

    :param op_types: Types of modules to track.
    :param graph: Graph to work with.
    :return: List of all first nodes with type in `op_types`.
    """
    graph_roots = graph.get_input_nodes()  # NNCFNodes here

    visited = {node_id: False for node_id in graph.get_all_node_ids()}
    partial_traverse_function = partial(traverse_function, type_check_fn=lambda x: x in op_types, visited=visited)

    first_nodes_of_type = []
    for root in graph_roots:
        first_nodes_of_type.extend(graph.traverse_graph(root, partial_traverse_function))
    return first_nodes_of_type


def get_split_axis(input_shapes: List[List[int]], output_shapes: List[List[int]]) -> int:
    """
    Returns split/chunk axis by given input and output shape of split/chunk node.

    :param input_shapes: Input_shapes of given split/chunk node.
    :param output_shapes: Output_shapes of given split/chunk node.
    :returns: Split/Chunk axis of given split/chunk node.
    """
    axis = None
    for idx, (dim_in, dim_out) in enumerate(zip(input_shapes[0], output_shapes[0])):
        if dim_in != dim_out:
            axis = idx
            break

    if axis is None:
        axis = -1
        nncf_logger.debug("Identity split/concat node detected")

    return axis


def get_number_of_quantized_ops(
    graph: NNCFGraph,
    quantizer_metatypes: List[Type[OperatorMetatype]],
    quantizable_metatypes: List[Type[OperatorMetatype]],
) -> int:
    """
    Returns the number of quantized operations in the graph.

    :param graph: NNCF graph which was built for the quantized model.
    :param quantizer_metatypes: List of quantizer's metatypes. For example,
        it may be metatype correspondent to FakeQuantize operation or
        metatypes correspondent to Quantize/Dequantize operations.
    :param quantizable_metatypes: List of metatypes for operations
        that may be quantized.
    :return: Number of quantized operations in the graph.
    """
    quantized_ops: Set[NNCFNode] = set()
    nodes_to_see: List[NNCFNode] = []

    for quantizer_node in graph.get_nodes_by_metatypes(quantizer_metatypes):
        nodes_to_see.extend(graph.get_next_nodes(quantizer_node))
        while nodes_to_see:
            node = nodes_to_see.pop()
            if node.metatype in quantizable_metatypes:
                quantized_ops.add(node)
            else:
                nodes_to_see.extend(graph.get_next_nodes(node))
    return len(quantized_ops)


def get_reduction_axes(
    channel_axes: Union[List[int], Tuple[int, ...]], shape: Union[List[int], Tuple[int, ...]]
) -> Tuple[int, ...]:
    """
    Returns filtered reduction axes without axes that correspond to channels.

    :param channel_axes: Channel axes.
    :param shape: Shape that need to be filtered.
    :return: Reduction axes.
    """
    reduction_axes = list(range(len(shape)))
    for channel_axis in sorted(channel_axes, reverse=True):
        del reduction_axes[channel_axis]
    return tuple(reduction_axes)



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import OrderedDict
from inspect import getfullargspec
from typing import Dict

import tensorflow as tf

import nncf
from nncf.tensorflow.layers.custom_objects import NNCF_CUSTOM_OBJECTS
from nncf.tensorflow.layers.custom_objects import get_nncf_custom_objects
from nncf.tensorflow.layers.operation import InputType
from nncf.tensorflow.layers.operation import NNCFOperation


@NNCF_CUSTOM_OBJECTS.register()
class NNCFWrapper(tf.keras.layers.Wrapper):
    """
    This wrapper augments a keras layer so the NNCF Operations may be applied to weights,
    callable attributes (like activations), input and output of the wrapped layer.
    """

    def __init__(self, layer, **kwargs):
        """
        Create a pruning wrapper for a keras layer.

        :param layer: the keras layer to be wrapped
        :param kwargs: additional keyword arguments to be passed to the keras layer.
        """
        if layer is None:
            raise ValueError("`layer` cannot be None.")

        if not isinstance(layer, tf.keras.layers.Layer) or isinstance(layer, tf.keras.Model):
            raise ValueError(
                "`layer` can only be a `tf.keras.layers.Layer` instance. "
                "You passed an instance of type: {input}.".format(input=layer.__class__.__name__)
            )

        if "name" not in kwargs:
            kwargs["name"] = layer.name

        super().__init__(layer, **kwargs)
        self._track_trackable(layer, name="layer")

        self.weights_attr_ops: Dict[str, Dict[str, NNCFOperation]] = {}

        self._init_layer_call_fn_args()
        self._trainable_weights = []
        self._non_trainable_weights = []
        self._ops_weights = {}
        self._op_build = False
        self._layer_weights = {}

    @property
    def trainable(self):
        return self.layer.trainable

    @trainable.setter
    def trainable(self, value):
        """
        By default, trainable property of operations follows the trainable property of the layer:
        - To freeze both wrapped layer weights and op weights -> layer.trainable = False
        - To unfreeze both wrapped layer weights and op weights -> layer.trainable = True
        """
        self.layer.trainable = value
        self.set_ops_trainable(value)

    def set_ops_trainable(self, value):
        """
        Introduction of trainable property to the operations gives the following additional options:
        - Frozen wrapped layer weights, unfrozen op weights -> layer.trainable = False; layer.set_ops_trainable(True)
        - Unfrozen wrapped layer weights, frozen op weights -> layer.trainable = True; layer.set_ops_trainable(False)
        """
        for ops in self.weights_attr_ops.values():
            for op in ops.values():
                op.trainable = value

    @property
    def trainable_weights(self):
        trainable_weights_by_definition = self._trainable_weights + self.layer.trainable_weights

        if self._op_build:
            trainable_ops_weights_on, trainable_ops_weights_off = self.classify_trainable_ops_weights()
        else:
            trainable_ops_weights_on, trainable_ops_weights_off = [], []

        if self.trainable:
            weights = trainable_weights_by_definition
            for trainable_weight_off in trainable_ops_weights_off:
                weights.remove(trainable_weight_off)
            outputs = weights
        else:
            outputs = trainable_ops_weights_on
        return outputs

    @property
    def non_trainable_weights(self):
        trainable_weights_by_definition = self._trainable_weights + self.layer.trainable_weights
        non_trainable_weights_by_definition = self._non_trainable_weights + self.layer.non_trainable_weights

        if self._op_build:
            trainable_ops_weights_on, trainable_ops_weights_off = self.classify_trainable_ops_weights()
        else:
            trainable_ops_weights_on, trainable_ops_weights_off = [], []

        if self.trainable:
            outputs = non_trainable_weights_by_definition + trainable_ops_weights_off
        else:
            weights = trainable_weights_by_definition
            for trainable_weight_on in trainable_ops_weights_on:
                weights.remove(trainable_weight_on)
            outputs = non_trainable_weights_by_definition + weights
        return outputs

    def classify_trainable_ops_weights(self):
        """
        Classifies operation trainable weights by trainable property of the corresponding operation.
        Note: Operation with trainable weights can be switched off from training.
              Operation trainable property can be adjusted
              to be different from NNCFWrapper global trainable property.

        :return trainable_ops_weights_on: List of trainable operation weights
                which is getting updated during training.
        :return trainable_ops_weights_off: List of trainable operation weights
                which is NOT getting updated during training.
        """
        trainable_ops_weights_on = []
        trainable_ops_weights_off = []
        for ops in self.weights_attr_ops.values():
            for op_name, op in ops.items():
                op_weights = self._ops_weights[op_name]
                for weight_val in op_weights.values():
                    if not weight_val.trainable:
                        continue

                    if op.trainable:
                        trainable_ops_weights_on.append(weight_val)
                    else:
                        trainable_ops_weights_off.append(weight_val)
        return trainable_ops_weights_on, trainable_ops_weights_off

    @property
    def updates(self):
        return self.layer.updates + self._updates

    @property
    def losses(self):
        return self.layer.losses + self._losses

    @property
    def data_format(self):
        return getattr(self.layer, "data_format", "channels_last")

    @property
    def ops_weights(self):
        return self._ops_weights

    @property
    def layer_weights(self):
        return self._layer_weights

    def build(self, input_shape=None):
        super().build(input_shape)
        for weight_attr, ops in self.weights_attr_ops.items():
            weight = self.get_layer_weight(weight_attr)
            for op_name, op in ops.items():
                self._ops_weights[op_name] = op.build(weight.shape, InputType.WEIGHTS, weight_attr, self)
            self._layer_weights[weight_attr] = weight
            self._trainable_weights.append(weight)
        self._op_build = True

    def call(self, inputs, training=None):
        training = self._get_training_value(training)

        self._apply_ops(training)

        if self._layer_expects_training_arg:
            outputs = self.layer.call(inputs, training=training)
        else:
            outputs = self.layer.call(inputs)

        return outputs

    def _apply_ops(self, training):
        for weight_attr, ops in self.weights_attr_ops.items():
            layer_weight = self._layer_weights[weight_attr]
            for op_name, op in ops.items():
                layer_weight = op(layer_weight, self._ops_weights[op_name], training)
            self.set_layer_weight(weight_attr, layer_weight)

    def registry_weight_operation(self, weights_attr: str, op: NNCFOperation):
        if weights_attr not in self.weights_attr_ops:
            self.weights_attr_ops[weights_attr] = OrderedDict()

        if op.name in self.weights_attr_ops[weights_attr]:
            raise nncf.InternalError(
                f"Attempt to apply an operation with the same name {op.name} on layer weight twice"
            )

        self.weights_attr_ops[weights_attr][op.name] = op

    def get_operation_weights(self, operation_name):
        return self._ops_weights[operation_name]

    def get_layer_weight(self, weight_attr):
        weight = getattr(self.layer, weight_attr, None)
        if weight is not None:
            return weight
        for w in self.layer.weights:
            if w.name.split(":")[0] == weight_attr:
                return w
        return None

    def set_layer_weight(self, weight_attr, weights):
        if hasattr(self.layer, weight_attr):
            setattr(self.layer, weight_attr, weights)
        else:
            self._layer_weights[weight_attr].assign(weights)

    def _init_layer_call_fn_args(self):
        call_full_argspec = getfullargspec(self.layer.call)
        call_fn_args = self._get_call_fn_args(call_full_argspec)
        self._layer_expects_training_arg = "training" in call_fn_args

    @staticmethod
    def _get_call_fn_args(call_full_argspec):
        all_args = call_full_argspec.args + call_full_argspec.kwonlyargs
        if all_args and all_args[0] == "self":
            return all_args[1:]
        return all_args

    @staticmethod
    def _get_training_value(training):
        if training is None:
            training = tf.keras.backend.learning_phase()
            if tf.is_tensor(training):
                training = tf.cast(training, tf.bool)
            else:
                training = bool(training)
        return training

    def get_config(self):
        config = super().get_config()

        weights_attr_ops = {}
        for weights_attr, ops in self.weights_attr_ops.items():
            weights_attr_ops[weights_attr] = []
            for op_name in ops:
                op_config = tf.keras.utils.serialize_keras_object(ops[op_name])
                weights_attr_ops[weights_attr].append(op_config)
        config["weights_attr_operations"] = weights_attr_ops
        return config

    @classmethod
    def from_config(cls, config, custom_objects=None):
        config = config.copy()

        weights_attr_ops_config = config.pop("weights_attr_operations")

        layer = tf.keras.layers.deserialize(config.pop("layer"), custom_objects=custom_objects)
        wrapper = cls(layer=layer, **config)

        for weights_attr, operations in weights_attr_ops_config.items():
            for op_config in operations:
                wrapper.registry_weight_operation(
                    weights_attr, tf.keras.layers.deserialize(op_config, custom_objects=get_nncf_custom_objects())
                )

        return wrapper



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Callable, Dict, List, Tuple, TypeVar

import torch

from nncf import NNCFConfig
from nncf.experimental.torch.nas.bootstrapNAS.elasticity.elasticity_controller import ElasticityController
from nncf.experimental.torch.nas.bootstrapNAS.elasticity.multi_elasticity_handler import SubnetConfig
from nncf.experimental.torch.nas.bootstrapNAS.training.model_creator_helpers import resume_compression_from_state
from nncf.torch.checkpoint_loading import load_state
from nncf.torch.model_creation import create_nncf_network
from nncf.torch.nncf_network import NNCFNetwork

TModel = TypeVar("TModel")
ValFnType = Callable[[NNCFNetwork, Any], Any]


class TrainedSuperNet:
    """
    An interface for handling pre-trained super-networks. This class can be used to quickly implement
    third party solutions for subnetwork search on existing super-networks.
    """

    def __init__(self, elastic_ctrl: ElasticityController, nncf_network: NNCFNetwork):
        """
        Initializes the super-network interface.

        :param elastic_ctrl: Elasticity controller to activate subnetworks
        :param nncf_network: NNCFNetwork that wraps the original PyTorch model.
        """
        self._m_handler = elastic_ctrl.multi_elasticity_handler
        self._elasticity_ctrl = elastic_ctrl
        self._model = nncf_network

    @classmethod
    def from_checkpoint(
        cls,
        model: TModel,
        nncf_config: NNCFConfig,
        supernet_elasticity_path: str,
        supernet_weights_path: str,
    ) -> "TrainedSuperNet":
        """
        Loads existing super-network weights and elasticity information, and creates the SuperNetwork interface.

        :param model: base model that was used to create the super-network.
        :param nncf_config: configuration used to create the super-network.
        :param supernet_elasticity_path: path to file containing state information about the super-network.
        :param supernet_weights_path: trained weights to resume the super-network.
        :return: SuperNetwork with wrapped functionality.
        """
        nncf_network = create_nncf_network(model, nncf_config)
        compression_state = torch.load(supernet_elasticity_path, map_location=torch.device(nncf_config.device))
        model, elasticity_ctrl = resume_compression_from_state(nncf_network, compression_state)
        model_weights = torch.load(supernet_weights_path, map_location=torch.device(nncf_config.device))
        load_state(model, model_weights, is_resume=True)
        elasticity_ctrl.multi_elasticity_handler.activate_maximum_subnet()
        return TrainedSuperNet(elasticity_ctrl, model)

    def get_search_space(self) -> Dict:
        """
        :return: dictionary with possible values for elastic configurations.
        """
        return self._m_handler.get_search_space()

    def get_design_vars_info(self) -> Tuple[int, List[int]]:
        """
        :return: number of possible values in subnet configurations and
        the number of possible values for each elastic property.
        """
        self._m_handler.get_design_vars_info()

    def eval_subnet_with_design_vars(self, design_config: List, eval_fn: ValFnType, **kwargs) -> Any:
        """

        :return: the value produced by the user's function to evaluate the subnetwork.
        """
        self._m_handler.activate_subnet_for_config(self._m_handler.get_config_from_pymoo(design_config))
        return eval_fn(self._model, **kwargs)

    def eval_active_subnet(self, eval_fn: ValFnType, **kwargs) -> Any:
        """
        :param eval_fn: user's function to evaluate the active subnetwork.
        :return: value of the user's function used to evaluate the subnetwork.
        """
        return eval_fn(self._model, **kwargs)

    def eval_subnet(self, config: SubnetConfig, eval_fn: ValFnType, **kwargs) -> Any:
        """
        :param config: subnetwork configuration.
        :param eval_fn: user's function to evaluate the active subnetwork.
        :return: value of the user's function used to evaluate the subnetwork.
        """
        self.activate_config(config)
        return self.eval_active_subnet(eval_fn, **kwargs)

    def activate_config(self, config: SubnetConfig) -> None:
        """
        :param config: subnetwork configuration to activate.
        """
        self._m_handler.activate_subnet_for_config(config)

    def activate_maximal_subnet(self) -> None:
        """
        Activates the maximal subnetwork in the super-network.
        """
        self._m_handler.activate_maximum_subnet()

    def activate_minimal_subnet(self) -> None:
        """
        Activates the minimal subnetwork in the super-network.
        """
        self._m_handler.activate_minimum_subnet()

    def get_active_config(self) -> SubnetConfig:
        """
        :return: the active configuration.
        """
        return self._m_handler.get_active_config()

    def get_macs_for_active_config(self) -> float:
        """
        :return: MACs of active subnet.
        """
        return self._m_handler.count_flops_and_weights_for_active_subnet()[0] / 2e6

    def export_active_subnet_to_onnx(self, filename: str = "subnet") -> None:
        """
        Exports the active subnetwork to ONNX format.

        :param filename: name of the output file.
        """
        self._elasticity_ctrl.export_model(f"{filename}.onnx")

    def get_config_from_pymoo(self, pymoo_config: List) -> SubnetConfig:
        """
        Converts a Pymoo subnetwork configuration into a SubnetConfig.

        :param pymoo_config: subnetwork configuration in Pymoo format.
        :return: subnetwork configuration in SubnetConfig format.
        """
        return self._m_handler.get_config_from_pymoo(pymoo_config)

    def get_active_subnet(self) -> NNCFNetwork:
        """
        :return: the nncf network with the current active configuration.
        """
        return self._model



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from copy import deepcopy
from typing import Optional

import torch
import torch.fx
from torch.ao.quantization.pt2e.duplicate_dq_pass import DuplicateDQPass
from torch.ao.quantization.pt2e.port_metadata_pass import PortNodeMetaForQDQ
from torch.ao.quantization.pt2e.utils import _disallow_eval_train
from torch.ao.quantization.pt2e.utils import _fuse_conv_bn_
from torch.ao.quantization.quantizer import Quantizer
from torch.fx import GraphModule
from torch.fx.passes.infra.pass_manager import PassManager

import nncf
from nncf.common.factory import NNCFGraphFactory
from nncf.common.logging import nncf_logger
from nncf.data import Dataset
from nncf.experimental.quantization.algorithms.post_training.algorithm import ExperimentalPostTrainingQuantization
from nncf.experimental.torch.fx.constant_folding import constant_fold
from nncf.experimental.torch.fx.quantization.quantizer.openvino_adapter import OpenVINOQuantizerAdapter
from nncf.experimental.torch.fx.quantization.quantizer.openvino_quantizer import OpenVINOQuantizer
from nncf.experimental.torch.fx.quantization.quantizer.torch_ao_adapter import TorchAOQuantizerAdapter
from nncf.experimental.torch.fx.transformations import QUANTIZE_NODE_TARGETS
from nncf.experimental.torch.fx.transformations import compress_post_quantize_transformation
from nncf.quantization.advanced_parameters import AdvancedBiasCorrectionParameters
from nncf.quantization.advanced_parameters import AdvancedSmoothQuantParameters
from nncf.quantization.advanced_parameters import RangeEstimatorParameters


def quantize_pt2e(
    model: torch.fx.GraphModule,
    quantizer: Quantizer,
    calibration_dataset: Dataset,
    subset_size: int = 300,
    fast_bias_correction: Optional[bool] = True,
    smooth_quant: bool = False,
    bias_correction_params: Optional[AdvancedBiasCorrectionParameters] = None,
    smooth_quant_params: Optional[AdvancedSmoothQuantParameters] = None,
    activations_range_estimator_params: Optional[RangeEstimatorParameters] = None,
    weights_range_estimator_params: Optional[RangeEstimatorParameters] = None,
    batchwise_statistics: Optional[bool] = None,
    fold_quantize: bool = True,
    do_copy: bool = False,
) -> torch.fx.GraphModule:
    """
    Applies post-training quantization to the torch.fx.GraphModule provided model
    using provided torch.ao quantizer.

    :param quantizer: Torch ao quantizer to annotate nodes in the graph with quantization setups
        to convey the desired way of quantization.
    :param subset_size: Size of a subset to calculate activations
        statistics used for quantization.
    :param fast_bias_correction: Setting this option to `False` enables a different
        bias correction method which is more accurate, in general, and takes
        more time but requires less memory. None disables the bias correction algorithm.
    :param smooth_quant: Setting this option to `True` enables the SmoothQuant algorithm.
    :param bias_correction_params: Contains advanced parameters for fine-tuning bias correction algorithm.
    :param smooth_quant_params: Contains advanced alpha parameters for SmoothQuant algorithm.
    :param activations_range_estimator_params: Contains parameters for estimating the range
        of activations of the model.
    :param weights_range_estimator_params: Contains parameters for estimating the range
        of weights of the model.
    :param batchwise_statistics: Determines whether quantizer statistics should be calculated
        for each item of the batch or for the entire batch, default is None, which means
        it set True if batch_size > 1 otherwise False.
    :param fold_quantize: Boolean flag for whether fold the quantize op or not. The value is True by default.
    :param do_copy: The copy of the given model is being quantized if do_copy == True,
        otherwise the model is quantized inplace. Default value is False.
    """
    nncf_logger.warning("This is an experimental feature and may change in the future without notice.")

    if subset_size < 1:
        raise nncf.ValidationError("Subset size must be positive.")

    batch_size = calibration_dataset.get_batch_size()
    if batchwise_statistics is None:
        batchwise_statistics = batch_size is not None and batch_size > 1

    original_graph_meta = model.meta

    if do_copy:
        model = deepcopy(model)

    _fuse_conv_bn_(model)
    if isinstance(quantizer, OpenVINOQuantizer):
        quantizer = OpenVINOQuantizerAdapter(quantizer)
    else:
        quantizer = TorchAOQuantizerAdapter(quantizer)

    # Call transform_prior_quantization before the NNCFGraph creation
    transformed_model = quantizer.transform_prior_quantization(model)

    quantization_algorithm = ExperimentalPostTrainingQuantization(
        quantizer=quantizer,
        subset_size=subset_size,
        fast_bias_correction=fast_bias_correction,
        smooth_quant=smooth_quant,
        bias_correction_params=bias_correction_params,
        smooth_quant_params=smooth_quant_params,
        activations_range_estimator_params=activations_range_estimator_params,
        weights_range_estimator_params=weights_range_estimator_params,
        batchwise_statistics=batchwise_statistics,
    )

    nncf_graph = NNCFGraphFactory.create(transformed_model)
    quantized_model = quantization_algorithm.apply(transformed_model, nncf_graph, dataset=calibration_dataset)

    # Magic. Without this call compiled model
    # is not preformant
    quantized_model = GraphModule(quantized_model, quantized_model.graph)

    if fold_quantize:
        if isinstance(quantizer, OpenVINOQuantizerAdapter):
            compress_post_quantize_transformation(quantized_model)
        else:
            constant_fold(quantized_model, _quant_node_constraint)

    pm = PassManager([DuplicateDQPass()])

    quantized_model = pm(quantized_model).graph_module
    pm = PassManager([PortNodeMetaForQDQ()])
    quantized_model = pm(quantized_model).graph_module

    quantized_model.meta.update(original_graph_meta)
    quantized_model = _disallow_eval_train(quantized_model)
    # Each transformation adds a duplicate tensor value to the model buffer.
    #  This step removes the duplicates tensor values from the buffer.
    quantized_model = GraphModule(quantized_model, quantized_model.graph)

    return quantized_model


def _quant_node_constraint(n: torch.fx.Node) -> bool:
    """If there is any pure ops between get_attr and quantize op they will be const propagated
    e.g. get_attr(weight) -> transpose -> quantize -> dequantize*
    (Note: dequantize op is not going to be constant propagated)

    This filter is added because we don't want to constant fold the things that are not
    related to quantization
    """
    return n.op == "call_function" and n.target in QUANTIZE_NODE_TARGETS



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import defaultdict
from typing import Dict, Set, Tuple

import networkx as nx

from nncf.common.graph import NNCFGraph
from nncf.common.graph import NNCFNode
from nncf.common.logging import nncf_logger
from nncf.common.quantization.structs import NonWeightQuantizerId
from nncf.torch.layers import NNCFConv2d
from nncf.torch.nncf_network import NNCFNetwork
from nncf.torch.quantization.algo import QuantizationController
from nncf.torch.quantization.precision_init.adjacent_quantizers import GroupsOfAdjacentQuantizers
from nncf.torch.quantization.structs import NonWeightQuantizerInfo


class BitwidthGraph:
    def __init__(
        self,
        algo_ctrl: QuantizationController,
        model: NNCFNetwork,
        groups_of_adjacent_quantizers: GroupsOfAdjacentQuantizers,
        add_flops=False,
    ):
        nncf_graph = model.nncf.get_graph()
        self._nx_graph = nncf_graph.get_graph_for_structure_analysis()
        if add_flops:
            flops_per_module = model.nncf.get_flops_per_module()

            flops_vs_node_group: Dict[int, Tuple[int, Set[NNCFNode]]] = defaultdict(set)
            for idx, module_node_name_and_flops in enumerate(flops_per_module.items()):
                module_node_name, flops = module_node_name_and_flops
                node_set = set(nncf_graph.get_op_nodes_in_scope(nncf_graph.get_scope_by_node_name(module_node_name)))
                flops_vs_node_group[idx] = (flops, node_set)

        grouped_mode = bool(groups_of_adjacent_quantizers)
        for node_key, node in nncf_graph.nodes.items():
            color = ""
            operator_name = node.node_type
            module = model.nncf.get_containing_module(node.node_name)
            if isinstance(module, NNCFConv2d):
                color = "lightblue"
                if module.groups == module.in_channels and module.in_channels > 1:
                    operator_name = "DW_Conv2d"
                    color = "purple"
                kernel_size = "x".join(map(str, module.kernel_size))
                operator_name += f"_k{kernel_size}"
                padding_values = set(module.padding)
                padding_enabled = len(padding_values) >= 1 and padding_values.pop()
                if padding_enabled:
                    operator_name += "_PAD"
                if add_flops:
                    matches = [
                        f_nodes_tpl for idx, f_nodes_tpl in flops_vs_node_group.items() if node in f_nodes_tpl[1]
                    ]
                    assert len(matches) == 1
                    flops, affected_nodes = next(iter(matches))
                    operator_name += f"_FLOPS:{str(flops)}"
                    if len(affected_nodes) > 1:
                        node_ids = sorted([n.node_id for n in affected_nodes])
                        operator_name += "(shared among nodes {})".format(
                            ",".join([str(node_id) for node_id in node_ids])
                        )
            operator_name += "_#{}".format(node.node_id)
            target_node_to_draw = self._nx_graph.nodes[node_key]
            target_node_to_draw["label"] = operator_name
            target_node_to_draw["style"] = "filled"
            if color:
                target_node_to_draw["color"] = color

        non_weight_quantizers = algo_ctrl.non_weight_quantizers
        bitwidth_color_map = {2: "purple", 4: "red", 8: "green", 6: "orange"}
        for quantizer_id, quantizer_info in non_weight_quantizers.items():
            self._paint_activation_quantizer_node(
                nncf_graph, quantizer_id, quantizer_info, bitwidth_color_map, groups_of_adjacent_quantizers
            )
        for wq_id, wq_info in algo_ctrl.weight_quantizers.items():
            nodes = [nncf_graph.get_node_by_name(tp.target_node_name) for tp in wq_info.affected_insertions]
            if not nodes:
                raise AttributeError(
                    "Failed to get affected nodes for quantized module node: {}".format(wq_id.target_node_name)
                )
            preds = [nncf_graph.get_previous_nodes(node) for node in nodes]
            wq_nodes = []
            for pred_list in preds:
                for pred_node in pred_list:
                    if "UpdateWeight" in pred_node.node_name:
                        wq_nodes.append(pred_node)
            assert len(wq_nodes) == 1

            node = wq_nodes[0]
            node_id = node.node_id
            key = nncf_graph.get_node_key_by_id(node_id)
            nx_node_to_draw_upon = self._nx_graph.nodes[key]
            quantizer = wq_info.quantizer_module_ref
            bitwidths = quantizer.num_bits
            nx_node_to_draw_upon["label"] = "WFQ_[{}]_#{}".format(quantizer.get_quantizer_config(), str(node_id))
            if grouped_mode:
                group_id_str = "UNDEFINED"
                group_id = groups_of_adjacent_quantizers.get_group_id_for_quantizer(wq_id)
                if group_id is None:
                    nncf_logger.debug(f"No group for weight quantizer for: {wq_id}")
                else:
                    group_id_str = str(group_id)
                nx_node_to_draw_upon["label"] += "_G" + group_id_str
            nx_node_to_draw_upon["color"] = bitwidth_color_map[bitwidths]
            nx_node_to_draw_upon["style"] = "filled"

    def _paint_activation_quantizer_node(
        self,
        nncf_graph: NNCFGraph,
        quantizer_id: NonWeightQuantizerId,
        quantizer_info: NonWeightQuantizerInfo,
        bitwidth_color_map: Dict[int, str],
        groups_of_adjacent_quantizers: GroupsOfAdjacentQuantizers,
    ):
        affected_insertion_points_list = quantizer_info.affected_insertions

        for target_point in affected_insertion_points_list:
            nncf_node_name = target_point.target_node_name
            nncf_node = nncf_graph.get_node_by_name(nncf_node_name)
            node_id = nncf_node.node_id

            input_port_id = target_point.input_port_id

            if input_port_id is None:
                # Post-hooking used for activation quantization
                # Currently only a single post-hook can immediately follow an operation
                succs = list(nncf_graph.get_next_nodes(nncf_node))
                assert len(succs) == 1
                target_nncf_node_key = nncf_graph.get_node_key_by_id(succs[0].node_id)
            else:
                # Pre-hooking used for activation quantization
                previous_nodes = nncf_graph.get_previous_nodes(nncf_node)
                target_node = None
                for prev_node in previous_nodes:
                    prev_edge = nncf_graph.get_nx_edge(prev_node, nncf_node)
                    if prev_edge[NNCFGraph.INPUT_PORT_ID_EDGE_ATTR] == input_port_id:
                        target_node = prev_node
                        break

                assert target_node is not None, "Could not find a pre-hook quantizer node for a specific input port!"
                target_nncf_node_id = target_node.node_id
                target_nncf_node_key = nncf_graph.get_node_key_by_id(target_nncf_node_id)

            activation_fq_node = self._nx_graph.nodes[target_nncf_node_key]
            bitwidth = quantizer_info.quantizer_module_ref.num_bits
            activation_fq_node["color"] = bitwidth_color_map[bitwidth]
            activation_fq_node["style"] = "filled"
            node_id = activation_fq_node[NNCFNode.ID_NODE_ATTR]

            activation_fq_node["label"] = "AFQ_[{}]_#{}".format(
                quantizer_info.quantizer_module_ref.get_quantizer_config(), str(node_id)
            )
            grouped_mode = bool(groups_of_adjacent_quantizers)
            if grouped_mode:
                group_id_str = "UNDEFINED"
                group_id = groups_of_adjacent_quantizers.get_group_id_for_quantizer(quantizer_id)
                if node_id is None:
                    nncf_logger.debug(f"No group for activation quantizer: {target_nncf_node_key}")
                else:
                    group_id_str = str(group_id)
                activation_fq_node["label"] += "_G" + group_id_str

    def get(self) -> nx.DiGraph:
        return self._nx_graph



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import weakref
from typing import Any, Dict, Optional, Union


class HookHandle:
    """
    A handle to remove a hook. Does not guarantee that
    hook_id is unique for the hooks_registry.
    """

    def __init__(self, hooks_registry: Dict[Any, Any], hook_id: str):
        """
        :param hooks_registry: A dictionary of hooks, indexed by hook `id`.
        :param hook_id: Hook id to use as key in the dictionary of hooks.
        """
        self.hooks_registry_ref: Optional[weakref.ReferenceType[Dict[Any, Any]]] = weakref.ref(hooks_registry)
        self._hook_id = hook_id

    @property
    def hook_id(self) -> Union[int, str]:
        """
        Key to use to retrieve handle from the registry.
        """
        return self._hook_id

    def remove(self) -> None:
        """
        Removes the corresponding operation from the registered hooks registry if it is possible.
        """
        if self.hooks_registry_ref is None:
            return

        hooks_registry = self.hooks_registry_ref()
        if hooks_registry is not None and self.hook_id in hooks_registry:
            del hooks_registry[self.hook_id]
            self.hooks_registry_ref = None


def add_op_to_registry(hooks_registry: Dict[Any, Any], op: Any) -> HookHandle:
    """
    Registers op into the hooks_registry and returns HookHandler instance.

    :param hooks_registry: A dictionary of hooks, indexed by hook `id`.
    :param op: Operation to set to registered hooks registry.
    :return: HookHandle that contains the registry of hooks and
        the id of operation to remove it from the registry.
    """
    if hooks_registry:
        hook_id = max(map(int, hooks_registry))
        hook_id_str = str(hook_id + 1)
    else:
        hook_id_str = "0"
    hooks_registry[hook_id_str] = op
    return HookHandle(hooks_registry, hook_id_str)



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from contextlib import contextmanager
from typing import Generator, Optional, TypeVar

from nncf.common.utils.backend import BackendType
from nncf.common.utils.backend import get_backend

# General categories
NNCF_CATEGORY = "nncf"

# Backend categories
NNCF_TF_CATEGORY = "nncf_tf"
NNCF_PT_CATEGORY = "nncf_pt"
NNCF_PT_FX_CATEGORY = "nncf_pt_fx"
NNCF_ONNX_CATEGORY = "nncf_onnx"
NNCF_OV_CATEGORY = "nncf_ov"

# Dynamic categories
MODEL_BASED_CATEGORY = "model_based"

CURRENT_CATEGORY: Optional[str] = None

TModel = TypeVar("TModel")


def _set_current_category(category: Optional[str]) -> None:
    global CURRENT_CATEGORY
    CURRENT_CATEGORY = category


def get_current_category() -> Optional[str]:
    return CURRENT_CATEGORY


def get_model_based_category(model: TModel) -> str:
    category_by_backend = {
        BackendType.ONNX: NNCF_ONNX_CATEGORY,
        BackendType.OPENVINO: NNCF_OV_CATEGORY,
        BackendType.TORCH: NNCF_PT_CATEGORY,
        BackendType.TENSORFLOW: NNCF_TF_CATEGORY,
        BackendType.TORCH_FX: NNCF_PT_FX_CATEGORY,
    }
    category = None
    if model is not None:
        model_backend = get_backend(model)
        category = category_by_backend[model_backend]

    return category


@contextmanager
def telemetry_category(category: Optional[str]) -> Generator[Optional[str], None, None]:
    previous_category = get_current_category()
    _set_current_category(category)
    yield category
    _set_current_category(previous_category)



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC
from abc import abstractmethod
from collections import deque
from typing import Any, Deque, Dict, List, Optional, Tuple, Union, cast

import numpy as np
from numpy.typing import NDArray

from nncf.common.tensor import NNCFTensor
from nncf.common.tensor import TensorType
from nncf.common.tensor_statistics.reduction import get_per_channel_history

ReductionAxes = Tuple[int, ...]


class TensorStatisticCollectorBase(ABC):
    """Collector estimate statistics at the quantization point based on the provided reduction shape."""

    def __init__(self, reduction_shape: Optional[ReductionAxes] = None, num_samples: Optional[int] = None):
        """
        Initializes Tensor Statistic Collector

        :param reduction_shape: Shape that defines tensor dimensions to reduce.
        :param num_samples: Maximum number of samples to collect.
        """
        self._reduction_shape = reduction_shape
        self._enabled = True
        self._collected_samples = 0
        self._num_samples = num_samples

    @property
    def num_samples(self) -> Union[int, None]:
        return self._num_samples

    def register_input(self, x: TensorType) -> TensorType:
        """Registers input tensor"""
        if not self._enabled:
            return x
        if self._num_samples is not None and self._collected_samples >= self._num_samples:
            return x
        if self._reduction_shape is None:
            self._reduction_shape = tuple(range(len(cast(NNCFTensor, x).shape)))
        self._register_input(x)
        self._collected_samples += 1
        return x

    @abstractmethod
    def _register_input(self, x: TensorType) -> None:
        pass

    def get_statistics(self) -> Any:
        """Returns collected statistics, if present."""
        if self._collected_samples == 0:
            raise StatisticsNotCollectedError()
        return self._get_statistics()

    @abstractmethod
    def _get_statistics(self) -> Any:
        pass

    def enable(self) -> None:
        self._enabled = True

    def disable(self) -> None:
        self._enabled = False

    def reset(self) -> None:
        """Resets all the statistics in the collector."""
        self._collected_samples = 0
        self._reset()

    @abstractmethod
    def _reset(self) -> None:
        pass

    def collected_samples(self) -> int:
        return self._collected_samples


class StatisticsNotCollectedError(Exception):
    """Raised when the statistics are not collected but requested."""


class OnlineTensorStatisticCollector(TensorStatisticCollectorBase):
    """Base class for collectors that collects statistics in online regime, without storing the data."""


class OfflineTensorStatisticCollector(TensorStatisticCollectorBase):
    """Collects statistics in offline regime by storing the data and aggregating it afterwards."""

    def __init__(
        self, reduction_shape: Optional[ReductionAxes] = None, num_samples: int = None, window_size: int = None
    ):
        super().__init__(reduction_shape, num_samples)
        self._samples: Deque[int] = deque(maxlen=window_size)

    def _reset(self) -> None:
        self._samples.clear()


class MinMaxStatisticCollector(OnlineTensorStatisticCollector):
    """Collector estimates min of minimum values and max of maximum values."""

    def __init__(self, use_abs_max: bool, reduction_shape: ReductionAxes, num_samples: int = None):
        super().__init__(reduction_shape, num_samples)
        self._use_abs_max = use_abs_max
        self._tensor_processor = self._get_processor()

        self._min_values = None
        self._max_values = None

    @staticmethod
    @abstractmethod
    def _get_processor() -> Any:
        pass

    def _register_input_common(self, x: NNCFTensor) -> None:
        min_reduced = self._tensor_processor.reduce_min(x, self._reduction_shape)
        if self._use_abs_max:
            x = self._tensor_processor.abs(x)
        max_reduced = self._tensor_processor.reduce_max(x, self._reduction_shape)

        if self._min_values is None:
            self._min_values = min_reduced
        else:
            self._min_values = self._tensor_processor.min(min_reduced, self._min_values)

        if self._max_values is None:
            self._max_values = max_reduced
        else:
            self._max_values = self._tensor_processor.max(max_reduced, self._max_values)

    def _reset(self) -> None:
        self._min_values = None
        self._max_values = None


class MinMaxOfflineStatisticCollectorBase(OfflineTensorStatisticCollector):
    """
    Base class for collectors that aggregate statistics
    from minimum and maximum values of tensors.
    """

    def __init__(
        self,
        use_per_sample_stats: bool,
        use_abs_max: bool,
        reduction_shape: ReductionAxes,
        num_samples: int = None,
        window_size: int = None,
    ):
        super().__init__(reduction_shape, num_samples)
        self._use_per_sample_stats = use_per_sample_stats
        self._use_abs_max = use_abs_max
        self._tensor_processor = self._get_processor()

        self._all_min_values: Deque[int] = deque(maxlen=window_size)
        self._all_max_values: Deque[int] = deque(maxlen=window_size)

    @staticmethod
    @abstractmethod
    def _get_processor() -> Any:
        pass

    def _register_input_common(self, x: NNCFTensor) -> None:
        min_reduced = self._tensor_processor.reduce_min(x, self._reduction_shape)
        if self._use_abs_max:
            x = self._tensor_processor.abs(x)
        max_reduced = self._tensor_processor.reduce_max(x, self._reduction_shape)

        if self._use_per_sample_stats:
            self._all_min_values.extend(self._tensor_processor.unstack(min_reduced))
            self._all_max_values.extend(self._tensor_processor.unstack(max_reduced))
        else:
            self._all_min_values.append(min_reduced)
            self._all_max_values.append(max_reduced)

    @abstractmethod
    def _min_aggregate(self) -> None:
        pass

    @abstractmethod
    def _max_aggregate(self) -> None:
        pass

    def _reset(self) -> None:
        self._all_min_values.clear()
        self._all_max_values.clear()


class MixedMinMaxStatisticCollector(MinMaxOfflineStatisticCollectorBase):
    """
    Collector aggregates (min or mean) of minimum values and (max or mean) of maximum values.
    """

    def __init__(
        self,
        use_per_sample_stats: bool,
        use_abs_max: bool,
        use_means_of_mins: bool,
        use_means_of_maxs: bool,
        reduction_shape: ReductionAxes,
        num_samples: int = None,
        window_size: int = None,
    ):
        super().__init__(use_per_sample_stats, use_abs_max, reduction_shape, num_samples, window_size)
        self._use_means_of_mins = use_means_of_mins
        self._use_means_of_maxs = use_means_of_maxs

    def _min_aggregate(self) -> Any:
        stacked_min = self._tensor_processor.stack(self._all_min_values)
        if self._use_means_of_mins:
            return self._tensor_processor.mean(stacked_min, axis=0)
        return self._tensor_processor.reduce_min(stacked_min, axis=0)

    def _max_aggregate(self) -> Any:
        stacked_max = self._tensor_processor.stack(self._all_max_values)
        if self._use_means_of_maxs:
            return self._tensor_processor.mean(stacked_max, axis=0)
        return self._tensor_processor.reduce_max(stacked_max, axis=0)


class MeanMinMaxStatisticCollector(MinMaxOfflineStatisticCollectorBase):
    """
    Collector aggregates mean of minimum values and mean of maximum values.
    """

    def _min_aggregate(self) -> Any:
        stacked_min = self._tensor_processor.stack(self._all_min_values)
        return self._tensor_processor.mean(stacked_min, axis=0)

    def _max_aggregate(self) -> Any:
        stacked_max = self._tensor_processor.stack(self._all_max_values)
        return self._tensor_processor.mean(stacked_max, axis=0)


class MeanStatisticCollector(OfflineTensorStatisticCollector):
    """
    Collector that aggregates statistics as mean along a pre-assigned axis.
    """

    def __init__(self, channel_axis: int, num_samples: Optional[int] = None, window_size: Optional[int] = None) -> None:
        """
        :param channel_axis: The main axis for the reduction while statistics collection.
        :param num_samples: Optional parameter for statistic collection that regulates
            the number of samples that will be processed.
        :param window_size: Optional maximum length for the statistic collection
        """
        super().__init__(num_samples=num_samples)
        self._channel_axis = channel_axis
        self._tensor_processor = self._get_processor()
        self._all_values: Deque[int] = deque(maxlen=window_size)
        self._all_shapes: Deque[int] = deque(maxlen=window_size)

    @staticmethod
    @abstractmethod
    def _get_processor() -> Any:
        pass

    def _register_input_common(self, x: NNCFTensor) -> None:
        if self._channel_axis == 0:
            self._all_values.append(self._tensor_processor.batch_mean(x))
        else:
            self._all_values.append(self._tensor_processor.mean_per_channel(x, self._channel_axis))
        self._all_shapes.append(cast(int, x.shape))

    def _reset(self) -> None:
        self._all_values.clear()
        self._all_shapes.clear()

    def _mean_aggregate(self) -> Any:
        all_values_stack = self._tensor_processor.stack(self._all_values)
        return self._tensor_processor.mean(all_values_stack, 0)

    def _shape(self) -> Any:
        return self._all_shapes[0]


class RawStatisticCollector(OfflineTensorStatisticCollector):
    """
    Collects tensor samples, where each tensor represented in raw format.
    Each sample stays available for usage in further stages of the algorithm.
    """

    def __init__(self, num_samples: Optional[int] = None) -> None:
        """
        :param num_samples: Optional parameter for statistic collection that regulates
            the number of samples that will be processed.
        """
        super().__init__(num_samples=num_samples)
        self._all_values: List[int] = []

    @staticmethod
    @abstractmethod
    def _get_processor() -> Any:
        pass

    def _register_input_common(self, x: NNCFTensor) -> None:
        self._all_values.append(cast(int, x.tensor))

    def _reset(self) -> None:
        self._all_values.clear()


class MedianMADStatisticCollector(OfflineTensorStatisticCollector):
    """
    Collector estimates median and median absolute deviation (MAD).
    """

    def _prepare_statistics(self) -> Tuple[NDArray[Any], NDArray[Any]]:
        per_channel_history = get_per_channel_history(
            self._samples, cast(List[int], self._reduction_shape), discard_zeros=True
        )
        per_channel_median = [np.median(channel_hist) for channel_hist in per_channel_history]
        per_channel_mad = []
        for idx, median in enumerate(per_channel_median):
            per_channel_mad.append(np.median(abs(per_channel_history[idx] - median)))
        numpy_median = np.asarray(per_channel_median)
        numpy_mad = np.asarray(per_channel_mad)
        return numpy_median, numpy_mad


class PercentileStatisticCollector(OfflineTensorStatisticCollector):
    """
    Collector estimates percentile values of all data history.
    """

    def __init__(
        self,
        percentiles_to_collect: List[float],
        reduction_shape: Optional[ReductionAxes] = None,
        num_samples: int = None,
        window_size: int = None,
    ):
        super().__init__(reduction_shape, num_samples, window_size)
        self._percentiles_to_collect = percentiles_to_collect

    def _prepare_statistics(self) -> Dict[float, Any]:
        per_channel_history = get_per_channel_history(self._samples, cast(List[int], self._reduction_shape))
        percentile_vs_values_dict = {}
        for pc in self._percentiles_to_collect:
            per_channel_percentiles = [np.percentile(channel_hist, pc) for channel_hist in per_channel_history]
            numpy_percentiles = np.asarray(per_channel_percentiles)
            percentile_vs_values_dict[pc] = numpy_percentiles
        return percentile_vs_values_dict


class MeanPercentileStatisticCollector(OfflineTensorStatisticCollector):
    """
    Collector estimates percentile values per step and then averages the results.
    """

    def __init__(
        self,
        percentiles_to_collect: List[float],
        reduction_shape: Optional[ReductionAxes] = None,
        num_samples: int = None,
        window_size: int = None,
    ):
        super().__init__(reduction_shape, num_samples, window_size)
        self._all_pct_values: Dict[float, Any] = {}
        for pc in percentiles_to_collect:
            self._all_pct_values[pc] = deque(maxlen=window_size)

    def _reset(self) -> None:
        for _, val in self._all_pct_values.items():
            val.clear()



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict

import torch

from nncf.common.factory import TModel
from nncf.common.graph.graph import NNCFGraph
from nncf.common.graph.transformations.commands import TransformationPriority
from nncf.common.graph.transformations.layout import TransformationLayout
from nncf.common.tensor_statistics.aggregator import StatisticPointsContainer
from nncf.common.tensor_statistics.aggregator import StatisticsAggregator
from nncf.common.utils.backend import BackendType
from nncf.experimental.common.tensor_statistics.statistics import TensorStatistic
from nncf.tensor import Tensor
from nncf.torch.graph.transformations.commands import PTInsertionCommand
from nncf.torch.graph.transformations.commands import PTTargetPoint
from nncf.torch.nncf_network import NNCFNetwork
from nncf.torch.tensor_statistics.algo import create_register_input_hook


class PTStatisticsAggregator(StatisticsAggregator):
    BACKEND: BackendType = BackendType.TORCH
    HOOKS_GROUP_NAME = "statistics_hooks"

    def collect_statistics(self, model: NNCFNetwork, graph: NNCFGraph) -> None:
        with torch.no_grad():
            super().collect_statistics(model, graph)
        model.nncf.remove_hooks_group(self.HOOKS_GROUP_NAME)

    def _register_statistics(self, outputs: Dict[str, Tensor], statistic_points: StatisticPointsContainer) -> None:
        # PyTorch backend doesn't use outputs to register statistics
        return

    def _get_transformation_layout_extra_outputs(
        self, statistic_points: StatisticPointsContainer
    ) -> TransformationLayout:
        transformation_layout = TransformationLayout()
        transformation_commands = []

        for _statistic_points in statistic_points.values():
            for _statistic_point in _statistic_points:
                for collectors in _statistic_point.algorithm_to_tensor_collectors.values():
                    for collector in collectors:
                        transformation_commands.append(
                            PTInsertionCommand(
                                _statistic_point.target_point,
                                create_register_input_hook(collector=collector),
                                TransformationPriority.FP32_TENSOR_STATISTICS_OBSERVATION,
                                hooks_group_name=self.HOOKS_GROUP_NAME,
                            )
                        )

        for transformation_command in transformation_commands:
            transformation_layout.register(transformation_command)

        return transformation_layout

    @staticmethod
    def _get_merged_statistic_points(
        statistic_points: StatisticPointsContainer, model: TModel, graph: NNCFGraph
    ) -> StatisticPointsContainer:
        # TODO: mirgate to experimental statistic collector and use common merging algorithm
        return statistic_points

    @staticmethod
    def _process_outputs(outputs: torch.Tensor) -> Dict[str, Tensor]:
        # PyTorch backend doesn't use outputs to register statistics
        return {}

    def _get_statistics_key(self, statistics: TensorStatistic, target_point: PTTargetPoint) -> str:
        """
        Returns key of statistics.

        :param statistics: Statistics value.
        :param target_point: Statistics target point.
        :return: Statistics key.
        """
        target_point_id = f"{target_point.target_node_name}_{target_point.type}_{target_point.input_port_id}"
        return f"{statistics.__class__.__name__}_{target_point_id}"



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from typing import Any, Dict

import jsonschema

from nncf.config.definitions import ALGO_NAME_VS_README_URL
from nncf.config.definitions import CONST_SPARSITY_ALGO_NAME_IN_CONFIG
from nncf.config.definitions import FILTER_PRUNING_ALGO_NAME_IN_CONFIG
from nncf.config.definitions import KNOWLEDGE_DISTILLATION_ALGO_NAME_IN_CONFIG
from nncf.config.definitions import MAGNITUDE_SPARSITY_ALGO_NAME_IN_CONFIG
from nncf.config.definitions import QUANTIZATION_ALGO_NAME_IN_CONFIG
from nncf.config.definitions import RB_SPARSITY_ALGO_NAME_IN_CONFIG
from nncf.config.definitions import SCHEMA_VISUALIZATION_URL
from nncf.config.schemata.accuracy_aware import ACCURACY_AWARE_MODES_VS_SCHEMA
from nncf.config.schemata.accuracy_aware import ACCURACY_AWARE_TRAINING_SCHEMA
from nncf.config.schemata.algo.const_sparsity import CONST_SPARSITY_SCHEMA
from nncf.config.schemata.algo.filter_pruning import FILTER_PRUNING_SCHEMA
from nncf.config.schemata.algo.knowledge_distillation import KNOWLEDGE_DISTILLATION_SCHEMA
from nncf.config.schemata.algo.magnitude_sparsity import MAGNITUDE_SPARSITY_SCHEMA
from nncf.config.schemata.algo.quantization import QUANTIZATION_SCHEMA
from nncf.config.schemata.algo.rb_sparsity import RB_SPARSITY_SCHEMA
from nncf.config.schemata.basic import ARRAY_OF_NUMBERS
from nncf.config.schemata.basic import BOOLEAN
from nncf.config.schemata.basic import STRING
from nncf.config.schemata.basic import make_object_or_array_of_objects_schema
from nncf.config.schemata.basic import with_attributes
from nncf.config.schemata.common.compression import COMPRESSION_LR_MULTIPLIER_PROPERTY
from nncf.config.schemata.defaults import TARGET_DEVICE
from nncf.config.schemata.experimental_schema import EXPERIMENTAL_REF_VS_ALGO_SCHEMA

logger = logging.getLogger("nncf")

REF_VS_ALGO_SCHEMA = {
    QUANTIZATION_ALGO_NAME_IN_CONFIG: QUANTIZATION_SCHEMA,
    FILTER_PRUNING_ALGO_NAME_IN_CONFIG: FILTER_PRUNING_SCHEMA,
    MAGNITUDE_SPARSITY_ALGO_NAME_IN_CONFIG: MAGNITUDE_SPARSITY_SCHEMA,
    RB_SPARSITY_ALGO_NAME_IN_CONFIG: RB_SPARSITY_SCHEMA,
    KNOWLEDGE_DISTILLATION_ALGO_NAME_IN_CONFIG: KNOWLEDGE_DISTILLATION_SCHEMA,
    CONST_SPARSITY_ALGO_NAME_IN_CONFIG: CONST_SPARSITY_SCHEMA,
    **EXPERIMENTAL_REF_VS_ALGO_SCHEMA,
}


SINGLE_INPUT_INFO_SCHEMA = {
    "type": "object",
    "properties": {
        "sample_size": with_attributes(
            ARRAY_OF_NUMBERS,
            description="Shape of the tensor expected as input to the model.",
            examples=[[1, 3, 224, 224]],
        ),
        "type": with_attributes(STRING, description="Data type of the model input tensor."),
        "filler": with_attributes(
            STRING,
            description="Determines what the tensor will be filled with when passed to the model"
            " during tracing and exporting.",
        ),
        "keyword": with_attributes(
            STRING,
            description="Keyword to be used when passing the tensor to the model's 'forward' method - "
            "leave unspecified to pass the corresponding argument as a positional arg.",
        ),
    },
    "additionalProperties": False,
}

TARGET_DEVICE_SCHEMA = {"type": "string", "enum": ["ANY", "CPU", "GPU", "NPU", "TRIAL", "CPU_SPR"]}


NNCF_CONFIG_SCHEMA = {
    "$schema": "http://json-schema.org/draft-07/schema",
    "title": "NNCF configuration file schema",
    "description": "The NNCF configuration file follows the JSON format and is the primary way to configure "
    "the result of NNCF application to a given user model. This configuration file is "
    "loaded into the `NNCFConfig` object by the user at runtime, after which the `NNCFConfig` "
    "is passed to the NNCF functions that perform actual compression or "
    "preparations for compression-aware training. \n\n"
    "The NNCF JSON configuration file is usually set up on a per-model, per-compression use case "
    "basis to contain:\n"
    "- a description of one or more compression algorithms to be applied to the model\n"
    "- the configuration parameters for each of the chosen algorithms\n"
    "- additional settings depending on the NNCF use case or integration scenario, e.g. specifying "
    "parameters for accuracy-aware training, or specifying model input shape for frameworks "
    "that do not have this data encapsulated in the model object in general such as PyTorch)\n"
    "and other parameters, the list of which may extend with the ongoing development of NNCF.\n\n"
    "This schema serves as a reference for users to write correct NNCF configuration files and each "
    "loaded NNCF configuration file into an `NNCFConfig` object is validated against it.",
    "type": "object",
    "properties": {
        "input_info": with_attributes(
            make_object_or_array_of_objects_schema(SINGLE_INPUT_INFO_SCHEMA),
            description="Describe the specifics of your model inputs here. "
            "This information is used to build the internal graph representation "
            "that is leveraged for proper compression functioning, and for "
            "exporting the compressed model to an executable format.\n"
            "If this field is unspecified, NNCF will try to deduce the input shapes and tensor types for the graph "
            "building purposes based on dataloader objects that are passed to compression algorithms by the user.",
        ),
        "target_device": with_attributes(
            TARGET_DEVICE_SCHEMA,
            description="The target device, the specificity of which will be taken into "
            "account while compressing in order to obtain the best "
            "performance for this type of device. The default 'ANY' means "
            "compatible quantization supported by any HW. Set "
            "this value to 'TRIAL' if you are going to use a custom "
            "quantization schema.",
            default=TARGET_DEVICE,
        ),
        "compression": make_object_or_array_of_objects_schema(
            {"oneOf": [{"$ref": f"#/$defs/{algo_name}"} for algo_name in REF_VS_ALGO_SCHEMA]}
        ),
        "accuracy_aware_training": with_attributes(
            ACCURACY_AWARE_TRAINING_SCHEMA,
            description="Options for the execution of the NNCF-powered "
            "'Accuracy Aware' training pipeline. The 'mode' "
            "property determines the mode of the accuracy-aware "
            "training execution and further available parameters.",
        ),
        # Validation of each separate compression description schema occurs in a separate step.
        # This is required for better user feedback, since holistic schema validation is uninformative
        # if there is an error in one of the compression configs.
        **COMPRESSION_LR_MULTIPLIER_PROPERTY,
        "disable_shape_matching": with_attributes(
            BOOLEAN,
            description="[Deprecated] Whether to enable strict input tensor "
            "shape matching when building the internal graph "
            "representation of the model. Set this to false if your "
            "model inputs have any variable dimension other than "
            "the 0-th (batch) dimension, or if any non-batch "
            "dimension of the intermediate tensors in your model "
            "execution flow depends on the input dimension, "
            "otherwise the compression will most likely fail.",
        ),
        "log_dir": with_attributes(STRING, description="Log directory for NNCF-specific logging outputs."),
    },
    "$defs": REF_VS_ALGO_SCHEMA,
}


def validate_single_compression_algo_schema(
    single_compression_algo_dict: Dict[str, Any], ref_vs_algo_schema: Dict[str, Any]
) -> None:
    """single_compression_algo_dict must conform to BASIC_COMPRESSION_ALGO_SCHEMA (and possibly has other
    algo-specific properties"""
    algo_name = single_compression_algo_dict["algorithm"]
    if algo_name not in ref_vs_algo_schema:
        raise jsonschema.ValidationError(
            f"Incorrect algorithm name - must be one of {str(list(ref_vs_algo_schema.keys()))}"
        )
    try:
        jsonschema.validate(single_compression_algo_dict, schema=ref_vs_algo_schema[algo_name])
    except jsonschema.ValidationError as e:
        e.message = (
            f"While validating the config for algorithm '{algo_name}' , got:\n"
            + e.message
            + f"\nRefer to the algorithm subschema definition at {SCHEMA_VISUALIZATION_URL}\n"
        )
        if algo_name in ALGO_NAME_VS_README_URL:
            e.message += (
                f"or to the algorithm documentation for examples of the configs: "
                f"{ALGO_NAME_VS_README_URL[algo_name]}"
            )
        raise e


def validate_accuracy_aware_training_schema(single_compression_algo_dict: Dict[str, Any]) -> None:
    """
    Checks accuracy_aware_training section.
    """
    jsonschema.validate(single_compression_algo_dict, schema=ACCURACY_AWARE_TRAINING_SCHEMA)
    accuracy_aware_mode = single_compression_algo_dict.get("mode")
    if accuracy_aware_mode not in ACCURACY_AWARE_MODES_VS_SCHEMA:
        raise jsonschema.ValidationError(
            "Incorrect Accuracy Aware mode - must be one of ({})".format(
                ", ".join(ACCURACY_AWARE_MODES_VS_SCHEMA.keys())
            )
        )
    try:
        jsonschema.validate(single_compression_algo_dict, schema=ACCURACY_AWARE_MODES_VS_SCHEMA[accuracy_aware_mode])
    except Exception as e:
        raise e



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from copy import deepcopy
from typing import List

import torch

from nncf import NNCFConfig
from nncf.api.compression import CompressionStage
from nncf.common.accuracy_aware_training.training_loop import ADAPTIVE_COMPRESSION_CONTROLLERS
from nncf.common.graph import NNCFNode
from nncf.common.initialization.batchnorm_adaptation import BatchnormAdaptationAlgorithm
from nncf.common.schedulers import StubCompressionScheduler
from nncf.common.sparsity.schedulers import SPARSITY_SCHEDULERS
from nncf.common.sparsity.statistics import LayerThreshold
from nncf.common.sparsity.statistics import MagnitudeSparsityStatistics
from nncf.common.statistics import NNCFStatistics
from nncf.common.utils.api_marker import api
from nncf.config.extractors import extract_algo_specific_config
from nncf.config.extractors import extract_bn_adaptation_init_params
from nncf.config.schemata.defaults import MAGNITUDE_SPARSITY_WEIGHT_IMPORTANCE
from nncf.config.schemata.defaults import SPARSITY_INIT
from nncf.config.schemata.defaults import SPARSITY_LEVEL_SETTING_MODE
from nncf.torch.algo_selector import PT_COMPRESSION_ALGORITHMS
from nncf.torch.compression_method_api import PTCompressionAlgorithmController
from nncf.torch.nncf_network import NNCFNetwork
from nncf.torch.sparsity.base_algo import BaseSparsityAlgoBuilder
from nncf.torch.sparsity.base_algo import BaseSparsityAlgoController
from nncf.torch.sparsity.base_algo import SparseModuleInfo
from nncf.torch.sparsity.collector import PTSparseModelStatisticsCollector
from nncf.torch.sparsity.layers import BinaryMask
from nncf.torch.sparsity.magnitude.functions import WEIGHT_IMPORTANCE_FUNCTIONS
from nncf.torch.sparsity.magnitude.functions import calc_magnitude_binary_mask


@PT_COMPRESSION_ALGORITHMS.register("magnitude_sparsity")
class MagnitudeSparsityBuilder(BaseSparsityAlgoBuilder):
    def create_weight_sparsifying_operation(self, target_module_node: NNCFNode, compression_lr_multiplier: float):
        return BinaryMask(target_module_node.layer_attributes.get_weight_shape())

    def _build_controller(self, model: NNCFNetwork) -> PTCompressionAlgorithmController:
        return MagnitudeSparsityController(model, self._sparsified_module_info, self.config)


@api()
@ADAPTIVE_COMPRESSION_CONTROLLERS.register("pt_magnitude_sparsity")
class MagnitudeSparsityController(BaseSparsityAlgoController):
    """
    Controller for the magnitude sparsity algorithm in PT.
    """

    def __init__(self, target_model: NNCFNetwork, sparsified_module_info: List[SparseModuleInfo], config: NNCFConfig):
        super().__init__(target_model, sparsified_module_info)
        self._config = config
        self._algo_config = extract_algo_specific_config(self._config, "magnitude_sparsity")
        params = self._algo_config.get("params", {})

        self._weight_importance_fn = WEIGHT_IMPORTANCE_FUNCTIONS[
            params.get("weight_importance", MAGNITUDE_SPARSITY_WEIGHT_IMPORTANCE)
        ]
        self._mode = params.get("sparsity_level_setting_mode", SPARSITY_LEVEL_SETTING_MODE)
        self._scheduler = None
        sparsity_init = self._algo_config.get("sparsity_init", SPARSITY_INIT)

        if self._mode == "global":
            scheduler_params = deepcopy(params)
            scheduler_params["sparsity_init"] = sparsity_init
            scheduler_cls = SPARSITY_SCHEDULERS.get(params.get("schedule", "polynomial"))
            self._scheduler = scheduler_cls(self, scheduler_params)
        else:
            self._scheduler = StubCompressionScheduler()

        self._bn_adaptation = None

        self.set_sparsity_level(sparsity_init)

    def statistics(self, quickly_collected_only: bool = False) -> NNCFStatistics:
        collector = PTSparseModelStatisticsCollector(self.model, self.sparsified_module_info)
        model_statistics = collector.collect()

        threshold_statistics = []
        if self._mode == "global":
            global_threshold = self._select_threshold(
                model_statistics.sparsity_level_for_layers, self.sparsified_module_info
            )

        module_name_to_sparsity_level_map = {
            s.name: s.sparsity_level for s in model_statistics.sparsified_layers_summary
        }
        for minfo in self.sparsified_module_info:
            if self._mode == "global":
                threshold = global_threshold
            else:
                sparsity_level_for_sparse_module = module_name_to_sparsity_level_map[minfo.module_node_name]
                threshold = self._select_threshold(sparsity_level_for_sparse_module, [minfo])

            threshold_statistics.append(LayerThreshold(minfo.module_node_name, threshold))

        target_sparsity_level = self.scheduler.current_sparsity_level if self._mode == "global" else None

        stats = MagnitudeSparsityStatistics(model_statistics, threshold_statistics, target_sparsity_level)

        nncf_stats = NNCFStatistics()
        nncf_stats.register("magnitude_sparsity", stats)
        return nncf_stats

    def freeze(self, freeze: bool = True):
        for layer in self.sparsified_module_info:
            layer.operand.frozen = freeze

    @property
    def compression_rate(self):
        return self.statistics().magnitude_sparsity.model_statistics.sparsity_level

    @compression_rate.setter
    def compression_rate(self, sparsity_level: float):
        self.freeze(False)
        self.set_sparsity_level(sparsity_level)
        self.freeze(True)

    def set_sparsity_level(
        self,
        sparsity_level,
        target_sparsified_module_info: SparseModuleInfo = None,
        run_batchnorm_adaptation: bool = False,
    ):
        if sparsity_level >= 1 or sparsity_level < 0:
            raise AttributeError(
                "Sparsity level should be within interval [0,1), actual value to set is: {}".format(sparsity_level)
            )
        if target_sparsified_module_info is None:
            target_sparsified_module_info_list = self.sparsified_module_info  # List[SparseModuleInfo]
        else:
            target_sparsified_module_info_list = [target_sparsified_module_info]
        threshold = self._select_threshold(sparsity_level, target_sparsified_module_info_list)
        self._set_masks_for_threshold(threshold, target_sparsified_module_info_list)

        if run_batchnorm_adaptation:
            self._run_batchnorm_adaptation()

    def _select_threshold(self, sparsity_level, target_sparsified_module_info_list):
        all_weights = self._collect_all_weights(target_sparsified_module_info_list)
        if not all_weights:
            return 0.0
        all_weights_tensor, _ = torch.cat(all_weights).sort()
        threshold = all_weights_tensor[int((all_weights_tensor.size(0) - 1) * sparsity_level)].item()
        return threshold

    def _set_masks_for_threshold(self, threshold_val, target_sparsified_module_info_list):
        for layer in target_sparsified_module_info_list:
            if not layer.operand.frozen:
                layer.operand.binary_mask = calc_magnitude_binary_mask(
                    layer.module.weight, self._weight_importance_fn, threshold_val
                )

    def _collect_all_weights(self, target_sparsified_module_info_list: List[SparseModuleInfo]):
        all_weights = []
        for minfo in target_sparsified_module_info_list:
            all_weights.append(self._weight_importance_fn(minfo.module.weight).view(-1))
        return all_weights

    def compression_stage(self) -> CompressionStage:
        if self._mode == "local":
            return CompressionStage.FULLY_COMPRESSED

        if self.scheduler.current_sparsity_level >= self.scheduler.target_level:
            return CompressionStage.FULLY_COMPRESSED
        if self.scheduler.current_sparsity_level == 0:
            return CompressionStage.UNCOMPRESSED
        return CompressionStage.PARTIALLY_COMPRESSED

    def _run_batchnorm_adaptation(self):
        if self._bn_adaptation is None:
            self._bn_adaptation = BatchnormAdaptationAlgorithm(
                **extract_bn_adaptation_init_params(self._config, "magnitude_sparsity")
            )
        self._bn_adaptation.run(self.model)



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch

from nncf.common.tensor import NNCFTensor


class PTNNCFTensor(NNCFTensor):
    """
    A realisation of torch tensors wrapper for common NNCF algorithms.
    """

    def __init__(self, tensor: torch.tensor):
        # In case somebody attempts to wrap
        # tensor twice
        if isinstance(tensor, self.__class__):
            tensor = tensor.tensor

        super().__init__(tensor)

    @property
    def device(self) -> torch.device:
        return self._tensor.device

    def is_empty(self) -> bool:
        return self.tensor.numel() == 0
